import tempfile

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import rasterio
from shapely.geometry import LineString, Polygon

from pywatemsedem.geo.valid import (
    PywatemsedemInputError,
    PywatemsedemTypeError,
    valid_exists,
    valid_linesvector,
    valid_pointvector,
    valid_polygonvector,
    valid_raster,
    valid_rasterlist,
    valid_vector,
    valid_vectorlist,
)


def dummy_vector_data(vct, geometry_type):
    """Generate dummy vector data and safe to file

    Parameters
    ----------
    vct: str
        File path to temporary vector file
    geometry_type: str
        Either 'Polygon' or 'Point' dummy data
    """
    df = pd.DataFrame(
        {"City": ["Buenos Aires", "Salzburg"], "Country": ["Argentina", "Austria"]}
    )

    if geometry_type == "Polygon":
        p1 = Polygon([(0, 0), (1, 0), (1, 1)])
        p2 = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        geometry = [p1, p2]
    elif geometry_type == "Point":
        p1 = [-34.58, -15.78]
        p2 = [-58.66, -47.91]
        geometry = gpd.points_from_xy(p1, p2)
    elif geometry_type == "LineString":
        p1 = LineString([(2, 0), (2, 4), (3, 4)])
        p2 = LineString([(2, 1), (4, 4), (2, 4)])
        geometry = [p1, p2]
    gdf = gpd.GeoDataFrame(df, geometry=geometry)
    gdf.to_file(vct)


def generate_raster(rst):
    """Generate dummy raster data and save to file

    Parameters
    ----------
    rst: str
        File path to temporary raster file
    """
    X, Y = np.meshgrid(np.linspace(-4.0, 4.0, 2), np.linspace(-3.0, 3.0, 2))
    Z = 10.0 * np.ones(X.shape)

    new_dataset = rasterio.open(
        rst,
        "w",
        driver="GTiff",
        height=Z.shape[0],
        width=Z.shape[1],
        count=1,
        dtype=Z.dtype,
        crs="+proj=latlong",
    )
    new_dataset.write(Z, 1)
    new_dataset.close()


def dummy_fun():
    """Dummy function for printing name errors"""


def test_valid():
    """Test valid seprate operators for
    :func:`pywatemsedem.geo.valid.valid_input`-decorator."""
    # test valid exists
    rst = tempfile.NamedTemporaryFile(suffix=".tif").name
    with pytest.raises(PywatemsedemInputError) as excinfo:
        valid_exists(rst, dummy_fun)
    assert "does not exist, cannot execute" in str(excinfo.value)

    # check if valid exists
    generate_raster(rst)
    assert valid_exists(rst, dummy_fun)

    # test valid raster
    assert valid_raster(rst, dummy_fun)

    # test feed wrong file format in valid raster
    log = tempfile.NamedTemporaryFile(suffix=".txt").name
    f = open(log, "w")
    f.write("test")
    f.close()
    with pytest.raises(PywatemsedemTypeError) as excinfo:
        valid_raster(log, dummy_fun)
    assert "The rasterio engine in pywatemsedem cannot open" in str(excinfo.value)

    # test if valid rasterlist not works when inputting wrong input
    assert valid_rasterlist([rst, rst], dummy_fun)

    with pytest.raises(PywatemsedemInputError) as excinfo:
        valid_rasterlist(rst, dummy_fun)
    assert "is not a valid list of rasters input," in str(excinfo.value)

    # test if polygonvector works
    vct = tempfile.NamedTemporaryFile(suffix=".shp").name
    dummy_vector_data(vct, "Polygon")
    assert valid_polygonvector(vct, dummy_fun)

    # test if pointvector works
    vct = tempfile.NamedTemporaryFile(suffix=".shp").name
    dummy_vector_data(vct, "Point")
    assert valid_pointvector(vct, dummy_fun)

    # test if linesvector works
    vct = tempfile.NamedTemporaryFile(suffix=".shp").name
    dummy_vector_data(vct, "LineString")
    assert valid_linesvector(vct, dummy_fun)

    # test if valid vectorlist not works when inputting wrong input
    assert valid_vectorlist([vct, vct], dummy_fun, "LineString")

    with pytest.raises(PywatemsedemInputError) as excinfo:
        valid_vectorlist(rst, dummy_fun, "LineString")
    assert "is not a valid list of vectors input," in str(excinfo.value)

    # test feed wrong file format in valid vector
    with pytest.raises(PywatemsedemTypeError) as excinfo:
        valid_vector(rst, dummy_fun, "Polygon")
    assert "The fiona engine in pywatemsedem cannot open" in str(excinfo.value)

    # test not known required type
    with pytest.raises(IOError) as excinfo:
        valid_vector(rst, dummy_fun, "test")
    assert "Geomtry type 'test' not known." in str(excinfo.value)

    # test feed wrong type of vector  (vct = line, see test above) to valid polygonshape
    with pytest.raises(PywatemsedemTypeError) as excinfo:
        valid_polygonvector(vct, dummy_fun)
    assert "Geometry input type" in str(excinfo.value)



import numpy as np

from pywatemsedem.grasstrips import scale_ktc_with_grass_strip_width, scale_ktc_zhang


def create_ktc_cnws(
    landuse_parcels,
    cfactor,
    mask,
    ktc_low,
    ktc_high,
    ktc_limit,
    sfolder,
    grass=None,
    correction_width=True,
):
    """Create KTC map for a given year, season.

    Parameters
    ----------
    # TODO
    """
    rst_mask = sfolder.scenario_folder / "mask.rst"
    mask.write(rst_mask)

    if cfactor is None:
        msg = "Unable to create KTC map, C-factor map is not known"
        raise FileNotFoundError(msg)
    if landuse_parcels is None:
        msg = "Unable to create KTC map, Prc map is not known"
        raise FileNotFoundError(msg)

    # reclass based on C-factor
    arr_ktc = np.where(cfactor <= ktc_limit, ktc_low, ktc_high)

    # give certain landuse classes an extremely high ktc-value
    arr_ktc = np.where(np.isin(landuse_parcels.arr, [-2, -1, -5]), 9999, arr_ktc)

    # set KTC outside modeldomain to zero
    # KTC_arr = np.where(self.catchm.binarr == 1, KTC_arr, 0)

    if correction_width is not None and grass is not None:

        grass._geodata = scale_ktc_gdf_grass_strips(
            grass.geodata,
            ktc_low,
            ktc_high,
        )
        arr_grass = grass.rasterize(
            rst_mask, landuse_parcels.rp.epsg, col="KTC", gdal=True
        )
        arr_ktc = np.where(
            np.logical_and(landuse_parcels.arr == -6, arr_grass != mask.rp.nodata),
            arr_grass,
            arr_ktc,
        )
        arr_ktc = np.where(
            np.logical_and(landuse_parcels.arr == -6, arr_grass == mask.rp.nodata),
            ktc_low,
            arr_ktc,
        )
    else:
        arr_ktc = np.where(landuse_parcels.arr == -6, ktc_low, arr_ktc)

    return arr_ktc, grass


def scale_ktc_gdf_grass_strips(gdf_grass_strips, ktc_low, ktc_high):
    """Scale the ktc values for grass strips in the grass strips dataframe

    Parameters
    ----------
    gdf_grass_strips: geopandas.GeoDataFrame
        Grass strips geodataframe, 'width' and 'scale_ktc' should be present as
        columns.
    ktc_low: float
    ktc_high: float
        See :func:`pywatemsedem.grasstrips.scale_ktc_linear` and
        :func:`pywatemsedem.grasstrips.scale_ktc_zhang`.

    Returns
    -------
    gdf_grass_strips: geopandas.GeoDataFrame
        Updated with the KTC value and the linear sediment trapping efficiency
        (STE_linear).
        For definition STE_linear,
        :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`.

    Notes
    -----
    Whether the kTC is scaled for the grass strips depends on the grass strip
    attribute "scale_ktc". If it is not scaled, the ktc_low-value is assigned.
    """
    if ("width" not in gdf_grass_strips.columns) or (
        "scale_ktc" not in gdf_grass_strips.columns
    ):
        msg = "Columns 'BREEDTE' and 'scale_ktc' should be in grass strips dataframe"
        return KeyError(msg)

    arr_ktc_grass, arr_ste = scale_ktc_with_grass_strip_width(
        gdf_grass_strips["width"].values,
        scale_ktc_zhang,
        ktc_high=ktc_high,
    )

    try:
        arr_scale_mask = np.array(gdf_grass_strips["scale_ktc"], dtype=bool)
    except TypeError:
        msg = (
            "Type of 'scale_ktc' column cannot be converted to boolean. Cannot "
            "mask 'KTC' column."
        )
        raise TypeError(msg)

    arr_ktc_grass[~arr_scale_mask] = ktc_low
    arr_ste[~arr_scale_mask] = np.nan

    gdf_grass_strips["KTC"] = arr_ktc_grass
    gdf_grass_strips["STE_linear"] = arr_ste

    return gdf_grass_strips



import warnings
from typing import List

import fiona
import numpy as np
from pyproj import CRS
from pyproj.exceptions import CRSError
from rasterio import Affine


class RasterProperties:
    """Raster properties class

    Pywatemsedem makes use of rasterio and gdal for loading, writing and processing
    rasters/vectors. A small class is implemented to easily switch between raster
    geographic references of gdal and rasterio, respectively names *gdal_profile*
    and *rasterio_profile*. Assuming you start with loading a raster with rasterio:

    .. code-block::

        from pywatemsedem.geo.utils import load_raster
        from pywatemsedem.geo.rasterproperties import RasterProperties
        arr, _, rasterio_profile = load_raster("test.tif")
        rp = RasterProperties.from_rasterio(rasterio_profile)
        rp.gdal_profile
        ...

    Note that ``rasterio_profile`` is of type dictionary, as such one can start from
    a profile definition in a dictionary format, for instance for gdal:

    .. code-block::

        from pywatemsedem.geo.utils import load_raster
        from pywatemsedem.geo.rasterproperties import RasterProperties
        gdal_profile = {
            "nodata": -9999.0,
            "epsg": "EPSG:31370",
            "res": 20.0,
            "minmax": [201620.0, 153880.0, 207500.0, 164060.0],
            "ncols": 294,
            "nrows": 509,
        }
        rp = RasterProperties.from_gdal(gdal_profile)
        rp.rasterio_profile
        ...

    In addition, a ``RasterProfile`` instance can be generated from known raster
    bounds and a raster resolution
    (this can be useful for generating raster geographical references when
    no rasterio or gdal profile definition is available):

    .. code-block::

        from pywatemsedem.geo.rasterproperties import RasterProperties

        bounds = [201620.0, 153880.0, 207500.0, 164060.0]
        resolution = 20
        nodata= -9999
        epsg = 31370
        rp = RasterProperties(bounds, resolution, nodata, epsg)

    Notes
    -----
    1. Current implementation support storing of raster properties, yet is does not aim
       to provide functionalities to adapt raster properties (as these functionalities
       are present in rasterio).
    2. Current implementation does not support tiled rasters. In addition,
       it only support bands-interleaving as only single-band raster are used.
    3. Definition interleaving: the way multiple bands of a raster are saved to the
       raster (e.g. pixel-based, line-based, band-based)
    4. The coordinate reference system (crs) is defined in EPSG.
    5. Note that dtype in gdal operation is typically derived from input raster
       dtype that is used to execute the gdal operation. As such dtype is not
       defined as a key in the gdal_profile property.
    """

    def __init__(
        self,
        bounds: List[float],
        resolution: int,
        nodata: float,
        epsg: int,
        driver: str = "GTiff",
    ):
        """Initialize a RasterProperties instance

        Parameters
        ----------
        bounds: list
            The **first** entry is **x_left**, the **second** entry is **y_lower**, the
            **third** entry is **x_right** and the **fourth** entry is **y_upper**
            [x_left, y_lower, x_right, y_upper].
        resolution: int
            Spatial resolution.
        nodata: float
            No data values.
        epsg: int
            EPSG code should be a numeric value, see https://epsg.io/.
        dtype: str, default "float64"
            Raster type
        driver : GTiff | RST | SAGA
            Name of GDAL driver, see https://gdal.org/drivers/raster/index.html; only
            GTiff | RST | SAGA are supported.
        """
        self._bounds = bounds
        self._resolution = resolution

        self._ncols = None
        self._nrows = None
        self._set_ncols_nrows()

        self._nodata = nodata

        self._check_epsg(epsg)
        self._epsg = epsg

        self._check_driver(driver)
        self._driver = driver

        self._xcoord = None
        self._ycoord = None

    @classmethod
    def from_gdal(cls, gdal_profile: dict):
        """Set RasterProperties from a gdal profile

        Parameters
        -----------
        gdal_profile: dict
        """
        if not {"res", "ncols", "nrows", "minmax", "nodata", "epsg"}.issubset(
            gdal_profile.keys()
        ):
            msg = "Function input is not a gdal profile instance!"
            raise IOError(msg)

        return cls(
            bounds=gdal_profile["minmax"],
            resolution=gdal_profile["res"],
            nodata=gdal_profile["nodata"],
            epsg=int(gdal_profile["epsg"].split(":")[1]),
        )

    @property
    def gdal_profile(self) -> dict:
        """Return gdal profile

        Returns
        -------
        gdal_profile: dict
            See definition in this function.
        """
        gdal_profile = dict()
        gdal_profile["nodata"] = self._nodata
        gdal_profile["epsg"] = f"EPSG:{self._epsg}"
        # gdal_profile["epsg"] = f"{self._epsg}"
        gdal_profile["res"] = self._resolution
        gdal_profile["minmax"] = self._bounds
        gdal_profile["ncols"] = self._ncols
        gdal_profile["nrows"] = self._nrows

        return gdal_profile

    @classmethod
    def from_rasterio(cls, rasterio_profile: dict, epsg: int = None):
        """Set RasterProperties with a rasterio profile

        Parameters
        -----------
        rasterio_profile: dict
        epsg: int, default None
            EPSG code should be a numeric value, see https://epsg.io/.
        """
        if not {"transform", "width", "height", "crs", "nodata"}.issubset(
            rasterio_profile.keys()
        ):
            msg = "Function input is not a rasterio profile instance!"
            raise IOError(msg)

        bounds = [
            rasterio_profile["transform"][2],
            rasterio_profile["transform"][5]
            - rasterio_profile["height"] * rasterio_profile["transform"][0],
            rasterio_profile["transform"][2]
            + rasterio_profile["width"] * rasterio_profile["transform"][0],
            rasterio_profile["transform"][5],
        ]
        if epsg is None:
            epsg = rasterio_profile["crs"].to_epsg()
        return cls(
            bounds=bounds,
            resolution=rasterio_profile["transform"][0],
            nodata=rasterio_profile["nodata"],
            epsg=epsg,
        )

    @property
    def rasterio_profile(self) -> dict:
        """Return rasterio profile

        Returns
        -------
        rasterio_profile: dict
            See definition in this function.
        """
        rasterio_profile = dict()
        rasterio_profile["driver"] = self._driver
        rasterio_profile["height"] = self._nrows
        rasterio_profile["width"] = self._ncols
        rasterio_profile["crs"] = CRS.from_epsg(self._epsg)
        rasterio_profile["transform"] = Affine(
            int(self._resolution),
            0,
            int(self._bounds[0]),
            0,
            -int(self._resolution),
            int(self._bounds[3]),
        )
        rasterio_profile["count"] = 1
        rasterio_profile["nodata"] = self._nodata
        if self._driver == "GTiff":
            rasterio_profile["compress"] = "deflate"

        return rasterio_profile

    @property
    def bounds(self) -> List[float]:
        """Raster boundary coordinates.

        Returns
        -------
        list : [x_left, y_lower, x_right, y_upper]
            The **first** entry is **x_left**, the **second** entry is **y_lower**, the
            **third** entry is **x_right** and the **fourth** entry is **y_upper**.
        """
        return self._bounds

    @property
    def resolution(self) -> int:
        """Raster resolution"""
        return self._resolution

    def _set_ncols_nrows(self):
        """Set the number of columns and rows based on the bounds and the resolution"""
        self._ncols = int(round((self._bounds[2] - self._bounds[0]) / self._resolution))
        self._nrows = int(round((self._bounds[3] - self._bounds[1]) / self._resolution))

    @property
    def nodata(self) -> float:
        """No data value used in raster."""
        return self._nodata

    @staticmethod
    def _check_epsg(epsg: int):
        """Check EPSG code

        Parameters
        ----------
        epsg: int
            EPSG code should be a numeric value, see https://epsg.io/.
        """
        if type(epsg) is not int:
            msg = f"EPSG-code '{epsg}' need to be an integer code."
            raise TypeError(msg)
        try:
            CRS.from_epsg(epsg)
        except CRSError:
            msg = f"EPSG-code '{epsg}' is an unknown epsg-code."
            raise CRSError(msg)

    @property
    def epsg(self) -> int:
        """int: EPSG code of the raster projection should be a numeric value,
        see https://epsg.io/"""
        return self._epsg

    @property
    def nrows(self) -> int:
        """Number of rows in the raster

        Returns
        -------
        int
        """
        return self._nrows

    @property
    def ncols(self) -> int:
        """Number of columns in the raster

        Returns
        -------
        int
        """
        return self._ncols

    @staticmethod
    def _check_driver(driver: str):
        """Check if the defined driver is known to pywatemsedem."""
        if driver not in ["GTiff", "RST", "SAGA"]:
            msg = f"Raster property driver '{driver}' not supported in pywatemsedem."
            raise IOError(msg)

    @property
    def driver(self) -> str:
        """Name of GDAL driver

        see https://gdal.org/drivers/raster/index.html;
        only GTiff | RST | SAGA are supported."""
        return self._driver

    @property
    def xcoord(self):
        """Returns 1D-vector array of x-coordinates

        Returns
        --------
        numpy.ndarray
        """
        return np.linspace(self.bounds[0], self.bounds[2], self.ncols)

    @property
    def ycoord(self):
        """Returns 1D-vector array of y-coordinates

        Returns
        --------
        numpy.ndarray
        """
        return np.linspace(self.bounds[1], self.bounds[3], self.nrows)


def get_bounds_from_vct(vct_catchment, resolution, n_pixels_buffer=5):
    """Get bounds from vector catchment.

    Parameters
    ----------
    vct_catchment: pathlib.Path
        Vector file of catchment.
    resolution: int
        Spatial resolution, see :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
    n_pixels_buffer: int, default 5
        Number of pixels that have to be taken into account to expand boundaries

    Returns
    -------
    bounds: list
        See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
    """
    with fiona.open(vct_catchment) as c:
        # Get spatial extent of catchment.
        bounds = c.bounds
        bounds = [
            round(bounds[0]) - n_pixels_buffer * resolution,
            round(bounds[1]) - n_pixels_buffer * resolution,
            round(bounds[2]) + n_pixels_buffer * resolution,
            round(bounds[3]) + n_pixels_buffer * resolution,
        ]

    return bounds


def synchronize_bounds(target, source, resolution):
    """Synchronize target geographical bounds of a raster with source bounds given a
     raster resolution.

    Bounds are defined as a list of  xmin, ymin, xmax and ymax-values of a raster. The
    boundaries leading to the smallest geographical model are selected as new
    boundaries.

    Parameters
    ----------
    target: list
        To check target boundaries, see
        :class:`pywatemsedem.geo.rasterproperties.RasterProperties`.
    source: pathlib.Path
        Source boundaries to which target boundaries have to be checked, see
        :class:`pywatemsedem.geo.rasterproperties.RasterProperties`.
    resolution: int
        See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`

    Returns
    -------
    bounds: list
        Updated bounds
    """
    check = False
    for i in range(4):
        if i < 2:  # xmin en ymin
            rest = (target[i] - source[i]) % resolution
            if rest != 0:
                check = True
                target[i] = target[i] - rest

        else:  # xmax en ymax
            rest = (target[i] - source[i - 2]) % resolution
            if rest != 0:
                check = True
                target[i] = target[i] - rest + resolution
    if check:
        msg = (
            f"Synchronizing x/y-axis model boundaries with input raster `{source.stem}`"
        )
        warnings.warn(msg)
    return target



from dataclasses import dataclass

import numpy as np

from pywatemsedem.geo.utils import write_arr_as_rst


@dataclass
class ParcelsLanduse:
    """Class to generate parcels land-use raster.

    This parcels land-use raster is used as raster input for WaTEM/SEDEM pywatemsedem.

    Parameters
    ----------
    landuse_core: numpy.ndarray
        Core landuse raster
    rivers: numpy.ndarray
        River raster (-1)
    water: numpy.ndarray
        Water raster (-3)
    infrastructure: numpy.ndarray
        Infrastructure (-2)
    mask: numpy.ndarray
        Mask array ???
    landuse_parcels: numpy.ndarray, default None
        Landuse parcels (-1, -2, -3, -4, -5, -6)
    parcels: numpy.ndarray, default None
        Pacels (>0)
    grass_strips: numpy.ndarray, default None
        Grass (-6)
    """

    landuse_core: np.ndarray
    river: np.ndarray
    water: np.ndarray
    infrastructure: np.ndarray
    mask: np.ndarray
    nodata: float
    landuse_parcels: np.ndarray = None
    parcels: np.ndarray = None
    grass_strips: np.ndarray = None

    def create_parcels_landuse_raster(self, scheme="Degerickx2015_adjusted"):
        """Create parcels landuse raster with a defined scheme.

        Parameters
        ----------
        scheme: str, default 'Degerickx2015_adjusted'

        Returns
        -------
        numpy.ndarray
            Parcels landuse raster
        """
        if scheme == "Degerickx2015_adjusted":
            return create_parcels_landuse_degerick2015(
                self.landuse_core,
                self.river,
                self.water,
                self.infrastructure,
                self.mask,
                self.nodata,
                landuse_parcels=self.landuse_parcels,
                parcels=self.parcels,
                grass_strips=self.grass_strips,
            )
        else:
            msg = f"Scheme '{scheme}' for creating parcels landuse raster is not known."
            raise NotImplementedError(msg)


def create_parcels_landuse_degerick2015(
    landuse_core,
    river,
    water,
    infrastructure,
    mask,
    nodata,
    landuse_parcels=None,
    parcels=None,
    grass_strips=None,
):
    """Create parcels landuse raster as with Degerickx and Van Den Broeck [1] scheme
    (version VPO).

    This schemes map rasters from top to bottom : rivers, water, infrastructure,
    grass strips, landuse from parcels data source, parcelsand finally pywatemsedem landuse.
    Incemental mapping to the parcels landuse raster is done by mapping pixels from the
    input raster to empty pixels in the landuse raster
    (see also :func:`pywatemsedem.scenario.map_input_array_on_array`). The pywatemsedem landuse
    raster is used to fill gaps.

    Parameters
    ----------
    landuse_core: numpy.ndarray
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    river: numpy.ndarray
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    water: numpy.ndarray
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    infrastructure: numpy.ndarray
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    mask: numpy.ndarray
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    landuse_parcels: numpy.ndarray, default None
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    parcels: numpy.ndarray, default None
        See :class:`pywatemsedem.scenario.ParcelsLanduse`
    grass_strips: numpy.ndarray, default None
        See :class:`pywatemsedem.scenario.ParcelsLanduse`

    Returns
    -------
    arr: numpy.ndarray
        Parcels landuse raster

    References
    ----------
    [1] Degerickx, J., Van Den Broeck, M., 2015. Handleiding CN_WS.KU Leuven, Leuven,
        Belgium.
    """

    arr = river.copy()
    if water is not None:
        arr = map_input_array_on_array(arr, water, nodata)
    if infrastructure is not None:
        arr = map_input_array_on_array(arr, infrastructure, nodata)
    if grass_strips is not None:
        arr = map_input_array_on_array(arr, grass_strips, nodata)
    if landuse_parcels is not None:
        arr = map_input_array_on_array(arr, landuse_parcels, nodata)
    if parcels is not None:
        # avoid too many parcel ids in integer 16 raster
        parcels = np.where(parcels != nodata, (parcels % 32757), nodata)
        arr = map_input_array_on_array(arr, parcels, nodata)
    if landuse_core is not None:
        arr = map_input_array_on_array(arr, landuse_core, nodata)

    arr = np.where(mask == 1, arr, 0).astype("float64")

    return arr


def map_input_array_on_array(arr, arr_input, nodata, method="only_empty"):
    """Map an input array on an other array with a method

    Parameters
    ----------
    arr: numpy.ndarray
        To map to array
    arr_input: numpy.ndarray
        To map array
    nodata: float
        No data value in arr_input array
    method: str, default "only_empty"
        Mappig method:

        - "only_empty": only map values from input array to empty (nodata) items in
          array.

    Returns
    -------
    numpy.ndarray
    """
    if arr_input.shape != arr.shape:
        msg = "Dimension to map array for are different from to map to array."
        raise ValueError(msg)
    if method == "only_empty":
        cond = arr == nodata
        return np.where(cond, arr_input, arr).copy()


def get_source_landuse(
    landuse,
    maxprc_id,
    rasterio_profile,
    arr_mask,
    tempfolder,
    tag,
):
    """Get landuse raster
    Parameters
    ----------
    maxprc_id : int
        Maximum perceelskaart id.
    gdal_profile : dict
        Rasterio profile dictionary holding all metadata for geotiff rasters.
    rasterio_profile : dict
        Gdal dictionary holding all metadata for idrisi rasters.
    arr_mask : numpy.ndarray
        Binaire array defining model domain (1 else 0).
    tempfolder : str or pathlib.Path
        Folder to write temp files to.
    tag : str
        Extra tag which is added to filename.
    """
    # landgebruik
    land_arr = np.where(landuse.arr == 10, maxprc_id, landuse.arr).astype("int16")
    nolanduse = np.logical_and(land_arr == rasterio_profile["nodata"], arr_mask == 1)
    land_arr = np.where(nolanduse, maxprc_id, land_arr).astype("int16")
    tmp_LandUse = tempfolder / f"landuse_reclass{tag}"
    write_arr_as_rst(land_arr, tmp_LandUse, "int16", rasterio_profile)

    return land_arr



import logging

import numpy as np

logger = logging.getLogger(__name__)


def process_buffer_outlets(gdf_outlets, gdf_buffer):
    """Map user-defined outlets to buffers.

    This function maps outlets to buffer id's based on the overlap between the outlet
    definition and the buffer polygon definition.

    #TODO: write test function

    Parameters
    ----------
    gdf_outlets: gpd.GeoDataFrame
        GeoDataFrame with a geometry definition.
    gdf_buffer: gpd.GeoDataFrame
        Holding buffer id's and geometry

    Returns
    -------
    gdf_outlets: gpd.GeoDataFrame
        Updated outlets with buf id assigedn
    """
    gdf_outlets["BUF_ID"] = 0
    # 1. per buffer kijken welke outlets er in liggen en het
    # buf_id toekennen aan deze outlet
    for b_id, geom in gdf_buffer[["BUF_ID", "geometry"]]:
        gdf_outlets["test"] = gdf_outlets.within(geom)
        mask = gdf_outlets.loc[gdf_outlets["test"]]
        if not mask.empty:
            msg = f"multiple buffer outlets present in buffer" f"with id {b_id}."
            msg += " only one, random is outlet chosen!"
            logger.warning(msg)
            gdf_outlets.loc[mask.index[0], "BUF_ID"] = b_id
        elif mask.empty:
            msg = f"no buffer outlet defined for buffer with id " f"{b_id}."
            logger.warning(msg)
        else:
            gdf_outlets.loc[mask.index[0], "BUF_ID"] = b_id

    return gdf_outlets


def filter_outlets_in_arr_extension_id(gdf, arr, arr_dtm, arr_outlet=None):
    """Map buffer outlet on extension id raster

    This function filters outlet id's to single outlet id, for every buffer. The
    function returns an array with single entries for every buffer id and
    multiple entries for every buffer extension id.

    The algorithm takes a single pixel that

    - corresponds to the pixel defined in arr_outlet (use-defined).

    OR

    - corresponds to the minimum dtm value in the extension of the buffer.

    Parameters
    ----------
    gdf: geopandas.GeoDataFrame
        Holding id's (column: 'buf_id' (int)), all id's should be present in arr and
        vice versa.
    arr: numpy.ndarray
        Array with buffer outlet (<16 385) and extension id's (16 385>=). This array
        can hold multiple entries for buffer outlet. All outlet id's should be present
        in gdf
    arr_dtm: numpy.ndarray
        DTM array, should have dimension of arr
    arr_outlet: numpy.ndarray
        Holds number of forced outlets, should have dimension of arr.

    Returns
    -------
    arr: numpy.ndarray
        Array with single entries for every buffer id and multiple entries for every
        buffer extension id.
    """
    un_outlets = np.unique(arr_outlet) if arr_outlet is None else []

    for row in gdf.itertuples():
        if row.buf_id in un_outlets:
            outlet_loc = arr_dtm[np.where(arr_outlet == row.buf_id)]
        else:
            outlet_loc = arr_dtm[arr == row.buf_exid].min()
        arr[(arr_dtm == outlet_loc) & (arr == row.buf_exid)] = row.buf_id
    return arr


def process_buffers_in_river(gdf, arr, arr_river, nodata):
    """This functions filters buffers from the buffer raster that overlap with a
    river

    Parameters
    ----------
    gdf: geopandas.GeoDataFrame
        Geopandas representation of arr, in buf_exid.
    arr: numpy.ndarray
        Array with buffer outlet (<16 385) and extension id's (16 385>=). This array
        can hold multiple entries for buffer outlet. All outlet id's should be present
        in gdf.
    arr_river: numpy.ndarray
        River raster, should be -1 and nodata
    nodata: float
        Nodata value in arr_river.

    Returns
    -------
    gdf: geopandas.GeoDataFrame
        Filtered
    arr: numpy.ndarray
        Filtered

    """
    # remove buffer pixels that coincide with river pixels
    arr = np.where(arr_river == -1, nodata, arr).astype("int16")
    # remove buffers that are not present in raster
    cond = np.isin(gdf["buf_exid"], np.unique(arr))
    gdf = gdf.loc[cond]

    return arr, gdf



from pathlib import Path

import numpy as np
import pandas as pd

from pywatemsedem.geo.utils import load_raster
from pywatemsedem.io.modeloutput import load_total_sediment_file


def _compare_folder(
    folder_benchmark,
    folder_output,
    year="2018",
    name="molenbeek",
    scenario="scenario_1",
):
    """Compare the setup of the folder/file structure

    Notes
    -----
    The filecmp, https://docs.python.org/3/library/filecmp.html, is not used
    as the file
    content can differ with a given precision on value-level. This functions
    checks
    which files/folders, specific functions are defined to test the data
    content.
    """

    # Check main folder is user defined name
    assert _get_subdir_names(folder_benchmark) == _get_subdir_names(folder_output)

    # Check data and catchment subfolders
    assert _get_subdir_names(folder_benchmark / name).issuperset(
        _get_subdir_names(folder_output / name)
    )
    assert {"Data_Bekken", scenario}.issubset(_get_subdir_names(folder_output / name))

    # Check scenario subfolders
    assert sorted(_get_subdir_names(folder_output / name / scenario)) == sorted(
        {str(year), "modelinput", "modeloutput", "postprocessing"}
    )

    # check 'modelinput' files
    assert sorted(
        _get_filenames(folder_benchmark / name / scenario / "modelinput")
    ) == sorted(_get_filenames(folder_output / name / scenario / "modelinput"))

    # check 'modeloutput' files
    assert sorted(
        _get_filenames(folder_benchmark / name / scenario / "modeloutput")
    ) == sorted(_get_filenames(folder_output / name / scenario / "modeloutput"))

    # check 'Data_bekken' files
    assert sorted(_get_filenames(folder_benchmark / name / "Data_Bekken")) == sorted(
        _get_filenames(folder_output / name / "Data_Bekken")
    )
    # check 'year' files
    assert sorted(
        _get_filenames(folder_benchmark / name / scenario / str(year))
    ) == sorted(_get_filenames(folder_output / name / scenario / str(year)))


def _compare_rst(
    folder_benchmark,
    folder_output,
    year="2018",
    name="molenbeek",
    scenario="scenario_1",
    ext=".rst",
    to_check_folders=["modelinput", "modeloutput"],
):
    """rst files in input and outfolder should be the same"""
    for folder in to_check_folders:
        _compare_rst_folder(
            folder_benchmark / name / scenario / folder,
            folder_output / name / scenario / folder,
            ext,
        )


def _compare_rst_folder(ref, new, ext=".rst"):
    """Compare RST files in two folders"""

    for file_name in ref.glob(f"*{ext}"):

        # adjust the numerical tolerance as function of the data; cumulative
        # data need to have more relative tolerance
        # TODO sgobeyn/daanr - revise these categories
        if file_name.name in [
            f"WATEREROS (mm per gridcel){ext}",
            f"RUSLE{ext}",
            f"Capacity{ext}",
            f"SediExport_kg{ext}",
            f"SediIn_kg{ext}",
            f"SLOPE{ext}",
        ]:
            rtol, atol = 1e-5, 1e-3
        elif file_name.name in [f"WATEREROS (kg per gridcel){ext}"]:
            rtol, atol = 1e-4, 1e-3
        elif file_name.name in [
            f"LS{ext}",
            f"AspectMap{ext}",
            f"sewer_in{ext}",
            f"cumulative{ext}",
            f"UPAREA{ext}",
            f"SediOut_kg{ext}",
        ]:
            rtol, atol = 1e-5, 1e-8
        else:
            rtol, atol = 1e-8, 1e-8
        equal_rst(
            file_name,
            (new / file_name.name),
            rtol=rtol,
            atol=atol,
        )


def _compare_table(
    folder_benchmark,
    folder_output,
    year="2018",
    name="molenbeek",
    scenario="scenario_1",
    to_check_folders=["modelinput", "modeloutput"],
):
    """rst files in input and outfolder should be the same"""

    for folder in to_check_folders:
        for file_name in (folder_benchmark / name / scenario / folder).glob("*.txt"):
            if file_name.name == "Total sediment.txt":
                equal_total_sediment(
                    file_name,
                    (folder_output / name / scenario / folder / file_name.name),
                    rtol=1e-8,
                    atol=1e-8,
                )
            elif file_name.name == "Total sediment VHA.txt":
                equal_table(
                    file_name,
                    (folder_output / name / scenario / folder / file_name.name),
                    rtol=1e-8,
                    atol=1e-8,
                    skiplines=1,
                )
            else:
                equal_table(
                    file_name,
                    (folder_output / name / scenario / folder / file_name.name),
                    rtol=1e-8,
                    atol=1e-8,
                    skiplines=0,
                )


def serialize_process_raster(rst):
    """Serializer function for rasters

    Parameters
    ----------
    rst: pathlib.path or str
        File path of to-assert raster

    Returns
    -------
    function:
        Return function that processes numpy array of pywatemsedem raster.
    """
    if "perceelskaart" in Path(rst).stem:
        return process_arr_cnwsperceelskaart
    else:
        return process_arr_default


def process_arr_cnwsperceelskaart(arr):
    """Remove unique parcels id's before comparing rasters

    Parameters
    ----------
    arr: np.ndarray
        WaTEM/SEDEM perceelskaart array to preprocess before moving to an assert
        statement.

    Returns
    -------
    arr: np.ndarray
        Processed WaTEM/SEDEM perceelskaart array to preprocess before moving to an
        assert statement.
    """
    arr[arr > 0] = 1
    return arr


def process_arr_default(arr):
    """Dummy function for processing arr

    Parameters
    ----------
    arr: np.ndarray
        Array to preprocess before moving to an assert statement.

    Returns
    -------
    arr: np.ndarray
        Equal to input.
    """
    return arr


def equal_ini(inifile_1, inifile_2, skip_lines=[1, 4, 5, 10]):
    """Check if settings.ini files are the same.

    Ignore difference in filepath and date
    lines [1, 4, 5, 10] contain paths and dates which are not compared

    Parameters
    ----------
    inifile_1: str or pathlib.Path
        File path of first ini-file.
    inifile_2: str or pathlib.Path
        File path of first ini-file.
    skip_lines: list
        Lines to skip in ini-file"""
    with open(inifile_1) as ini_1:
        with open(inifile_2) as ini_2:
            ini_content_1 = [
                line
                for idx, line in enumerate(ini_1.readlines())
                if idx not in skip_lines
            ]
            ini_content_2 = [
                line
                for idx, line in enumerate(ini_2.readlines())
                if idx not in skip_lines
            ]
            assert ini_content_1 == ini_content_2
            assert set(ini_content_1).difference(set(ini_content_2)) == set()


def equal_table(txt_table_file_1, txt_table_file_2, rtol=1e-8, atol=1e-8, skiplines=0):
    """Check if table data are the same

    Parameters
    ----------
    txt_table_file_1: str or pathlib.Path
        File path of first table-like file.
    txt_table_file_2: str or pathlib.Path
        File path of second table-like file.
    rtol: float
        See :func:`numpy.testing.assert_allclose`.
    atol: float
        See :func:`numpy.testing.assert_allclose`.
    skiplines: int
        See :func:`pandas.read_csv`.
    """
    tbl_1 = pd.read_csv(txt_table_file_1, skiprows=skiplines, delimiter="\t")
    tbl_2 = pd.read_csv(txt_table_file_2, skiprows=skiplines, delimiter="\t")
    pd.testing.assert_frame_equal(tbl_1, tbl_2, rtol=rtol, atol=atol)


def equal_total_sediment(
    txt_total_sediment_file_1, txt_total_sediment_file_2, rtol=1e-8, atol=1e-8
):
    """Check if total sediment output data of two files are the same

    Parameters
    ----------
    txt_total_sediment_file_1: str or pathlib.Path
        File path to WaTEM/SEDEM total sediment output file 1.
    txt_total_sediment_file_2: str or pathlib.Path
        File path to WaTEM/SEDEM total sediment output file 2.
    rtol: float
        See :func:`numpy.testing.assert_allclose`.
    atol: float
        See :func:`numpy.testing.assert_allclose`.
    """
    sed1 = load_total_sediment_file(txt_total_sediment_file_1)
    sed2 = load_total_sediment_file(txt_total_sediment_file_2)
    for val1, val2 in zip(sed1.values(), sed2.values()):
        assert np.allclose(val1, val2, rtol=rtol, atol=atol)


def equal_rst(rst_file_1, rst_file_2, rtol=1e-8, atol=1e-8):
    """Check if content of raster files are equal.

    Parameters
    ----------
    rst_file_1: pathlib.path
        File path of first raster.
    rst_file_2: pathlib.path
        File path of  second raster
    rtol: float
        See :func:`numpy.testing.assert_allclose`.
    atol: float
        See :func:`numpy.testing.assert_allclose`.
    """
    condition, _, _ = check_equal_rst(rst_file_1, rst_file_2, rtol, atol)

    if not condition:
        msg = f"Content {rst_file_1.stem} not equal to content " f"{rst_file_2.stem}."
        raise AssertionError(msg)


def check_equal_rst(rst_file_1, rst_file_2, rtol=1e-8, atol=1e-8):
    """Check if content of raster files are equal.

    Parameters
    ----------
    rst_file_1: pathlib.path
        File path of first raster.
    rst_file_2: pathlib.path
        File path of  second raster
    rtol: float
        See :func:`numpy.testing.assert_allclose`.
    atol: float
        See :func:`numpy.testing.assert_allclose`.

    """
    arr1, profile = load_raster(rst_file_1)
    arr2, profile = load_raster(rst_file_2)

    condition = np.allclose(arr1, arr2, rtol=rtol, atol=atol, equal_nan=True)

    return condition, arr1, arr2


def _get_subdir_names(path):
    """"""
    return set([item.name for item in path.iterdir() if item.is_dir()])


def _get_filenames(path):
    """"""
    return set(
        [
            item.name
            for item in path.iterdir()
            if (item.suffix in [".tif", ".rst", ".shp"])
        ]
    )


def _compare_cnws_input_raster(
    rst,
    rst_benchmark,
    tag,
    rtol=1e-8,
    atol=1e-8,
):
    """Compare two WaTEM/SEDEM rasters

    Check if they are equal, and if not, raise an Assertion with relevant statistics
    about difference. The statistics are dependent on the substansive content of the
    file.


    Parameters
    ----------
    rst: str or pathlib.Path
        File path of raster from test simulation
    rst_benchmark: str or pathlib.Path
        File path of raster from benchmark simulation
    rtol: float
    atol: float
    """
    condition, arr, arr_benchmark = check_equal_rst(rst, rst_benchmark, rtol, atol)

    if not condition:

        if tag.lower() in ["perceelskaart"]:
            out = compute_difference_table_cnws_prckrt_rasters(arr, arr_benchmark)
        elif tag.lower() in ["c-factor", "ktc"]:
            out = compute_difference_table_cnws_rasters(arr, arr_benchmark, tag)
        elif tag.lower() in ["buffers"]:
            out = compute_difference_table_cnws_buffers(arr, arr_benchmark)
        else:
            msg = f"``tag`` {tag} not known."
            raise IOError(msg)

        msg = (
            f"WaTEM/SEDEM '{tag}' rasters are not equal. Difference statistics table: "
            f"\n {out}"
        )
        raise AssertionError(msg)


def compute_difference_table_cnws_buffers(arr, arr_benchmark, buf_idext=2**14):
    """Compute table whcih indicates difference between two WaTEM/SEDEM buffer raster
    files.


    Parameters
    ----------
    arr: np.ndarray
    arr_benchmark: np.ndarray
    buf_idext: float
        Threshold value that differentiates in buffer raster which is a buffer_id and
        which is a buffer extension id, SEE XXX

    Returns
    -------
    str:
        String format of difference dataframe table.
    """
    n_buffers = np.unique(arr[arr <= buf_idext]).shape[0]
    n_buffers_benchmark = np.unique(arr_benchmark[arr_benchmark <= buf_idext]).shape[0]

    n_buffers_exid = np.unique(arr[arr > buf_idext]).shape[0]
    n_buffers_exid_benchmark = np.unique(
        arr_benchmark[arr_benchmark > buf_idext]
    ).shape[0]

    return pd.DataFrame(
        data=[
            [n_buffers, n_buffers_benchmark],
            [n_buffers_exid, n_buffers_exid_benchmark],
        ],
        columns=["test", "benchmark"],
        index=["n_buf_id", "n_buf_exid"],
    ).to_string()


def compute_difference_table_cnws_rasters(arr, arr_benchmark, tag):
    """Compute table whcih indicates difference between two WaTEM/SEDEM raster files.

    Parameters
    ----------
    arr: np.ndarray
    arr_benchmark: np.ndarray

    Returns
    -------
    str:
        String format of difference dataframe table.
    """
    arr_diff = arr - arr_benchmark

    return (
        pd.DataFrame(arr_diff[arr_diff != 0], columns=[f"diff_{tag}"])
        .describe()
        .to_string()
    )


def compute_difference_table_cnws_prckrt_rasters(
    arr,
    arr_benchmark,
    dict_translation={
        -6: "grass strips",
        -5: "pool",
        -4: "meadow",
        -3: "forest",
        -2: "infrastructure",
        -1: "river",
        0: "outside_domain",
        1: "agriculture",
    },
    res=20,
):
    """Compute table whcih indicates difference between two WaTEM/SEDEM
    perceelskaart raster files.

    Parameters
    ----------
    arr: np.ndarray
    arr_benchmark: np.ndarray

    Returns
    -------
    str:
        String format of difference dataframe table.
    """
    df = pd.DataFrame(columns=["left", "right"])
    df["test"] = arr.ravel()
    df["benchmark"] = arr_benchmark.ravel()
    df.loc[df["test"] > 0, "test"] = 1
    df.loc[df["benchmark"] > 0, "benchmark"] = 1
    df = df.replace(dict_translation)
    df["area (ha)"] = res**2 / 100**2
    df = df.groupby(["test", "benchmark"]).aggregate({"area (ha)": np.sum})

    return df.sort_values("area (ha)", ascending=False).to_string()



ALLOWED_RASTER_FORMATS = ["idrisi", "tiff"]

PREFIX_TEMP = "pywatemsedem_"

# fix this
SUFFIXES_TIF = [".tif.aux.xml", ".tif"]
SUFFIXES_SHP = [".shp", ".cpg", ".shx", ".prj", ".dbf", ".qix"]
SUFFIXES_SAGA = [
    ".sgrd",
    ".sdat.aux.xml",
    ".mgrd",
    ".sgrid",
    ".sdat",
    ".prj",
]  # TODO: why is .prj added?
SUFFIXES_RST = [".rst", ".rdc", ".rst.aux.xml"]
SUFFIXES_TXT = [".mtab", ".txt"]

# flags used for saga_cmd
SAGA_FLAGS = "-f=q"  #: no progress report



import datetime
import logging

# Standard libraries
import pathlib
from pathlib import Path

import pandas as pd

from pywatemsedem.tools import check_courant_criterium, get_item_from_ini

logger = logging.getLogger(__name__)


class UserChoices:
    def __init__(self):
        self.version = None
        self.dict_ecm_options = {}
        self.dict_model_options = {}
        self.dict_variables = {}
        self.dict_output = {}

    def set_ecm_options(self, defaultini=None):
        """Set the erosion control options for the scenario.

        The possible keys are:

        - 'Include buffers': use bufferbasins
        - 'UseGras': use grasstrips
        - 'Include dams': use dams
        - 'Include ditches': use ditches
        - 'UseTeelttechn': use culture specific measures (like strip till,...)

        Parameters
        ----------
        defaultini : str or pathlib.Path, optional
            ini-file with section 'ECMOptions' and one or more of the above
            mentioned keys as parameters
        """
        lst_mandatory_keys = [
            "Include buffers",
            "UseGras",
            "Include dams",
            "Include ditches",
            "UseTeelttechn",
        ]

        for key in lst_mandatory_keys:
            if key not in self.dict_ecm_options:
                # read key from default values
                if defaultini is not None:
                    defaultvalue = get_item_from_ini(
                        defaultini, "ECMOptions", key, "int"
                    )
                    msg = (
                        f"{key} is not given in the ECM options, "
                        f"default value ({defaultvalue}) is used"
                    )
                    logger.warning(msg)
                else:
                    defaultvalue = None

                if defaultvalue is not None:
                    self.dict_ecm_options[key] = defaultvalue
                else:
                    msg = f"No defaultvalue for ECMOption {key} given"
                    raise ValueError(msg)

    def set_model_version(self, version):
        """Set the model version to be used in the scenario.

        The model-version will be stored in the variables.

        Parameters
        ----------
        version: str
            'CN-WS' or 'WS' or 'Only Routing'
        """
        if version == "WS":
            self.version = version
            self.dict_variables["Only WS"] = 1
        elif version == "CN-WS":
            self.version = version
            self.dict_variables["Only WS"] = 0
        elif version == "Only Routing":
            self.version = "Only Routing"
            self.dict_variables["Only WS"] = 1
        else:
            msg = "No correct model version is chosen, please select 'WS' of 'CN-WS'"
            raise ValueError(msg)

    def set_transport_capacity_equation(self, TC="VanOost2000"):
        """Set the transport capacity equation model for the scenario.

        The chosen values will be stored in the ``ModelOptions``.

        Parameters
        ----------
        TC: 'VanOost2000' or 'Verstraeten2007', optional, default 'VanOost2000'
            Name of the transport capacity equation model.
        """
        self.set_model_equation("TC model", TC, ["VanOost2000", "Verstraeten2007"])

    def set_LS_equation(self, L="Desmet1996_Vanoost2003", S="Nearing1997"):
        """Set the L and S-model for the scenario.

        The chosen values will be stored in the ``ModelOptions``.

        Parameters
        ----------
        L: 'Desmet1996_Vanoost2003' or 'Desmet1996_McCool', optional, default
            'Desmet1996_Vanoost2003'
            Name of the L model.
        S: 'Nearing1997' or 'McCool1987', optional, default 'Nearing1997'
            Name of the S model.
        """
        self.set_model_equation(
            "L model", L, ["Desmet1996_Vanoost2003", "Desmet1996_McCool"]
        )
        self.set_model_equation("S model", S, ["Nearing1997", "McCool1987"])

    def set_model_equation(self, type_model, model_equation, lst_possible_equations):
        """Set the equation for a specific model according to a reference

        Parameters
        ----------
        type_model: str
            Name type of model.
        model_equation: str
            Specific reference for model equation, e.g. L according to
            "Desmet1996_Vanoost2003".
        lst_possible_equations: list
            Possible references to equation

        Note
        ----
        See example use in :func:`pywatemsedem.CNWS.UserChoices.set_LS_equation`.
        """
        if model_equation in lst_possible_equations:
            self.dict_model_options[type_model] = model_equation
        else:
            msg = (
                f"No correct {type_model} was given! please choose "
                f"{', '.join(lst_possible_equations)}."
            )
            raise KeyError(msg)

    def set_model_options(self, defaultini=None):
        """Set the ModelOptions for the scenario.

        The possible keys are:

        - 'Convert output'
        - 'CreateKTIL'
        - 'Estimate clay content'
        - 'Manual outlet selection'
        - 'Calibrate'
        - 'Adjusted Slope'
        - 'Buffer reduce Area'
        - 'Force Routing'
        - 'FilterDTM'
        - 'River Routing'

        When the modelversion is 'WS' one can also define 'UseR' as a key.

        Parameters
        ----------
        defaultini : str or pathlib.Path, optional
            ini-file with section 'Options' and one or more of the above
            mentioned keys as parameters
        """

        lst_mandatory_keys = [
            "Convert output",
            "Manual outlet selection",
            "Include sewers",
            "UserProvidedKTC",
            "Adjusted Slope",
            "Buffer reduce Area",
            "Force Routing",
            "FilterDTM",
            "River Routing",
            "OnlyInfraSewers",
            "Maximize grass strips",
        ]

        lst_extra_keys = [
            "Estimate clay content",
            "Calculate Tillage Erosion",
            "Create ktil map",
            "Calibrate",
        ]

        # if self.version == 'WS':
        #    lst_mandatory_keys.append('UseR')

        if self.version != "Only Routing":
            lst_mandatory_keys = lst_mandatory_keys + lst_extra_keys

        for key in lst_mandatory_keys:
            if key not in self.dict_model_options:
                if defaultini is not None:
                    # read key from default values
                    defaultvalue = get_item_from_ini(defaultini, "Options", key, "int")
                else:
                    defaultvalue = None

                if defaultvalue is not None:
                    msg = (
                        f"`{key}` is not given in the model options, default "
                        f"value `{defaultvalue}` is used."
                    )
                    logger.warning(msg)
                    self.dict_model_options[key] = defaultvalue
                else:
                    msg = f"No default value for Model Option {key} given"
                    raise ValueError(msg)

        if self.version == "Only Routing":
            for key in lst_extra_keys:
                self.dict_model_options[key] = 0

        return

    def set_model_variables(self, defaultini=None):
        """Set the ModelVariables for the scenario.

        Parameters
        ----------
        defaultini : str or pathlib.Path, optional
            ini-file with section 'Variables' and one or more of the above
            mentioned keys as parameters

        Notes
        -----
        The possible keys are: 'Bulk density', 'ktc limit',
        'Parcel connectivity cropland', 'Parcel connectivity forest',
        'Parcel trapping efficiency cropland', 'Parcel trapping efficiency forest',
        'Parcel trapping efficiency pasture', 'begin_jaar', 'begin_maand',
        'Max kernel', 'Max kernel river'/

        When the CN-module is used, additional variables are needed: 'Alpha',
        'Beta', 'Stream velocity', 'CNTable'.

        When Rainfalldata is used (i.e. WS and UseR=0 or CNWS), additional variables
        are needed: '5-day antecedent rainfall', 'Desired timestep for model',
        'Endtime model', 'Final timestep output', 'Rainfall filename'.

        If no Rainfalldata is used (i.e. WS using a general R-factor) add the 'R'
        variable.

        When calibrating the WS model, define 'KTcHigh_lower', 'KTcHigh_upper',
        'KTcLow_lower', 'KTcLow_upper' and 'steps'.

        Without calibrating, define 'ktc low', 'ktc high'.

        When the ModelOption['CreateKTIL'] is 1, define 'ktil default' and
        'ktil threshold'.

        When the ModelOption['Estimate clay content'] is 1, define
        'Clay content parent material'.

        When the ModelOption['Manual outlet selection']  is 1, define 'Outletshp'.

        When the ModelOption['Force Routing'] is 1, define 'Routingshp'.
        """
        lst_mandatory_keys = [
            "Parcel connectivity cropland",
            "Parcel connectivity forest",
            "Parcel connectivity grasstrips",
            "Parcel trapping efficiency cropland",
            "Parcel trapping efficiency forest",
            "Parcel trapping efficiency pasture",
            "begin_jaar",
            "begin_maand",
            "Max kernel",
            "Max kernel river",
            "Bulk density",
            "LS correction",
        ]

        lst_cn_keys = ["Alpha", "Beta", "Stream velocity"]

        lst_rainfall_keys = [
            "5-day antecedent rainfall",
            "Desired timestep for model",
            "Endtime model",
            "Final timestep output",
        ]

        if self.dict_model_options["Calibrate"] == 1:
            lst_ktc_keys = [
                "KTcHigh_lower",
                "KTcHigh_upper",
                "KTcLow_lower",
                "KTcLow_upper",
                "steps",
                "ktc limit",
            ]
        else:
            lst_ktc_keys = ["ktc low", "ktc high", "ktc limit"]

        lst_mandatory_keys += lst_ktc_keys

        if self.dict_model_options["Create ktil map"] == 1:
            lst_ktil_keys = ["ktil default", "ktil threshold"]
            lst_mandatory_keys += lst_ktil_keys

        if self.version == "CN-WS":
            lst_mandatory_keys += lst_cn_keys
            lst_mandatory_keys += lst_rainfall_keys

        lst_mandatory_keys.append("R factor")

        if self.dict_model_options["Estimate clay content"] == 1:
            lst_mandatory_keys.append("Clay content parent material")

        # if self.dict_model_options["Manual outlet selection"] == 1:
        #    lst_mandatory_keys.append("Outletshp")

        # if self.dict_model_options["Force Routing"] == 1:
        #    lst_mandatory_keys.append("Routingshp")

        if self.dict_model_options["Include sewers"] == 1:
            lst_sewer_keys = ["SewerInletEff", "Sewer exit"]
            lst_mandatory_keys += lst_sewer_keys

        for key in lst_mandatory_keys:
            if key not in self.dict_variables:
                if defaultini is not None:
                    if key in [
                        "ktc limit",
                        "ktc low",
                        "ktc high",
                        "ktil threshold",
                        "Alpha",
                        "Beta",
                        "Stream velocity",
                        "LS correction",
                    ]:
                        T = "float"
                    elif key in [
                        "Outletshp",
                        "Routingshp",
                    ]:
                        T = "get"
                    else:
                        T = "int"

                    # read key from default values
                    defaultvalue = get_item_from_ini(defaultini, "Variables", key, T)
                    msg = (
                        f"`{key}` is not given in the model variables, default"
                        f" value ({defaultvalue}) is used."
                    )
                    logger.warning(msg)
                else:
                    defaultvalue = None

                if defaultvalue is not None:
                    self.dict_variables[key] = defaultvalue
                else:
                    msg = f"No value for variable {key} given in default values"
                    raise ValueError(msg)
        return

    def set_output(self, defaultini=None):
        """Set the modeloutput for the scenario.

        The possible keys are:
        'Write aspect', 'Write LS factor', 'Write RUSLE', 'Write sediment export',
        'Write slope', 'WriteTotalRunof', 'Write upstream area', 'Write water erosion'

        Parameters
        ----------
        defaultini : str or pathlib.Path, optional
            ini-file with section 'Output' and one or more of the above
            mentioned keys as parameters
        """

        lst_mandatory_keys = [
            "Write aspect",
            "Write LS factor",
            "Write slope",
            "Write upstream area",
            "Output per river segment",
            "Write routing table",
        ]

        lst_ws_keys = ["Write RUSLE", "Write sediment export", "Write water erosion"]
        lst_cn_keys = ["Write rainfall excess", "Write total runoff"]

        if self.version == "WS" or self.version == "CN-WS":
            lst_mandatory_keys = lst_mandatory_keys + lst_ws_keys
        else:  # Only Routing
            for key in lst_ws_keys:
                self.dict_output[key] = 0

        if self.version == "CN-WS":
            lst_mandatory_keys += lst_cn_keys
        else:  # WS en Only Routing
            for key in lst_cn_keys:
                self.dict_output[key] = 0

        for key in lst_mandatory_keys:
            if key not in self.dict_output:
                if defaultini is not None:
                    # read key from default values
                    defaultvalue = get_item_from_ini(defaultini, "Output", key, "int")
                    msg = (
                        f"`{key}` is not given in the model options, default "
                        f"value ({defaultvalue}) is used."
                    )
                    logger.warning(msg)
                else:
                    defaultvalue = None

                if defaultvalue is not None:
                    self.dict_output[key] = defaultvalue
                else:
                    msg = (
                        f"No value for Model Output {key} given in default " f"values."
                    )
                    raise ValueError(msg)
        return

    def set_userchoices_curvenumber(
        self,
        type_cn_storage,
        txt_rainfall_filename,
        rainfall_depth_antecedent,
        begin_date,
        end_date,
        time_step_event,
        days_antecedent,
        resolution,
    ):
        """Set userchoices for Curve-Number model.

        Parameters
        ----------
        type_cn_storage: str
            "filestorage" or "database" storage.
        txt_rainfall_filename: str
            See :func:`pywatemsedem.CNWS.UserChoicesCurveNumber.check`
        rainfall_depth_antecedent: int
            Antecedent rainfall depth (mm).
        begin_date: pandas.DateTime
            See :func:`pywatemsedem.userchoices.UserChoicesCurveNumber.check`
        end_date: pandas.DateTime
            See :func:`pywatemsedem.userchoices.UserChoicesCurveNumber.check`
        time_step_event: int
            See :func:`pywatemsedem.userchoices.UserChoicesCurveNumber.check`
        days_antecedent: int
            See :func:`pywatemsedem.userchoices.UserChoicesCurveNumber.check`
        resolution: int
            Spatial resolution.
        """
        if txt_rainfall_filename is not None:
            type_cn_storage = "filestorage"

        userchoices_curvenumber = UserChoicesCurveNumber(type_cn_storage, resolution)
        (
            self.txt_rainfall_filename,
            self.rainfall_depth_antecedent,
            self.begin_date,
            self.end_date,
            self.days_antecedent,
            self.time_step_event,
        ) = userchoices_curvenumber.check(
            txt_rainfall_filename,
            rainfall_depth_antecedent,
            begin_date,
            end_date,
            time_step_event,
            days_antecedent,
            stream_velocity=self.dict_variables["Stream velocity"],
        )
        self.dict_variables["Desired timestep for model"] = self.time_step_event


class UserChoicesCurveNumber:
    def __init__(self, type_cn_storage, resolution):

        self.type_cn_storage = type_cn_storage
        self.resolution = resolution

    def check(
        self,
        txt_rainfall_filename=None,
        rainfall_depth_antecedent=None,
        begin_date=None,
        end_date=None,
        time_step_event=None,
        days_antecedent=None,
        stream_velocity=None,
    ):
        """Set the settings for an event based on whether a filesystem or database
        storage system is used. In a filebased system, a reference to a rainfall
        data file has to be given, with an amount of antecedent rainfall. In the
        other case the begin and end date have to be given, along with the days of
        antecedent, and the time step of an event.

        Parameters
        ----------
        txt_rainfall_filename: pathlib.Path or str, optional
            File path to rainfall input data file. This file is a headerless text
            file with two columns. The first column contains the time in minutes
            (starting from 0), the second column contains the rainfall in mm.
        rainfall_depth_antecedent: int or float
            Cumulative amount of rainfall in days_antecedent.
        begin_date: pandas.Timestamp, optional
            Begin rainfall timerseries, in UTC, make use of pd.Timestamp, see Note 1.
        end_date: pandas.Timestamp, optional
            End rainfall timerseries, in UTC, make use of pd.Timestamp, see Note 1.
        time_step_event: int
            Time step event.
        days_antecedent: int or float, optional
            Number of days for antecedent
        stream_velocity: float
            Stream velocity (m/s)

        Returns
        -------
        See `Parameters`.

        Note
        ----
        1. Use of begin_date_event and end_date_event is for example, with UTC as
        timezone:

        .. code-block:: python

            begin_date = pandas..to_datetime("2018-01-01", utc=True)
            end_date = pandas.to_datetime("2018-01-15", utc=True)

        """
        self.rainfall_filename = txt_rainfall_filename
        self.rainfall_depth_antecedent = rainfall_depth_antecedent

        if not isinstance(begin_date, pd.Timestamp) or not isinstance(
            end_date, pd.Timestamp
        ):
            msg = (
                "Begin date and end date should be a pandas.Timestamp object in UTC, "
                "use `begin_date = pd.to_datetime('YYYY-MM-DD' utc=True)`."
            )
            raise TypeError(msg)
        if (str(begin_date.tz) == "UTC") and (str(end_date.tz) == "UTC"):
            self.begin_date = begin_date
            self.end_date = end_date
        else:
            msg = (
                "Use timezone-UTC to define begin and end date, "
                "use `begin_date = pd.to_datetime('YYYY-MM-DD' utc=True)`."
            )
            raise IOError(msg)

        self.days_antecedent = days_antecedent
        self.time_step_event = time_step_event

        if self.type_cn_storage == "filestorage":

            self.check_choices_cn_filestorage()

        elif self.type_cn_storage == "databasestorage":

            self.check_choices_cn_databasestorage()

        if stream_velocity is not None:
            self.time_step_event = check_courant_criterium(
                stream_velocity, self.time_step_event, self.resolution
            )

        return (
            self.rainfall_filename,
            self.rainfall_depth_antecedent,
            self.begin_date,
            self.end_date,
            self.days_antecedent,
            self.time_step_event,
        )

    def check_choices_cn_filestorage(self):
        """Set the settings for an event for a filesystem storage system is used."""
        self.begin_date = None
        self.end_date = None
        self.days_antecedent = None

        if not (
            isinstance(self.rainfall_filename, str)
            or isinstance(self.rainfall_filename, pathlib.Path)
        ):
            msg = "Rainfall filename is not of type `str` or `pathlib.Path`."
            raise TypeError(msg)

        if not (
            isinstance(self.rainfall_depth_antecedent, int)
            or isinstance(self.rainfall_depth_antecedent, float)
        ):
            msg = "Rainfall depth antecedent is not or type `int` or `float`."
            raise TypeError(msg)

        if not Path(self.rainfall_filename).exists():
            msg = f"Rainfall filename {self.rainfall_filename} does not exist."
            raise IOError(msg)

    def check_choices_cn_databasestorage(self):
        """Set the settings for an event for a database storage system is used."""
        self.rainfall_filename = None
        self.rainfall_depth_antecedent = None

        if not (isinstance(self.begin_date, datetime.datetime)):
            msg = "Begin date is not of type `datetime.datetime`."
            raise TypeError(msg)

        if not (isinstance(self.end_date, datetime.datetime)):
            msg = "End date is not of type `datetime.datetime`."
            raise TypeError(msg)

        if not (self.end_date > self.begin_date):
            msg = "End date of the event should be later that begin date of the event."
            raise ValueError(msg)

        if not (
            isinstance(self.days_antecedent, int)
            or isinstance(self.days_antecedent, float)
        ):
            msg = "Days antecedent is not of type `float` or `int`."
            raise TypeError(msg)

        if not (
            isinstance(self.time_step_event, int)
            or isinstance(self.time_step_event, float)
        ):
            msg = "Time step event is not of type `float` or `int`."
            raise TypeError(msg)



from conftest import calibratedata

from pywatemsedem import calibrate


class TestCalibrate:
    """Test class to test calibrate functionlaities."""

    # TODO: expand tests with dummy data.
    def test_process_calibrationrun_output(self):
        """Test processing of calibration output"""
        # implicit test
        calibrate.process_calibrationrun_output(
            calibratedata.txt_calibrate,
            1000,
            "test_catch",
            50 * 1000**2,
            endpoint_coefficient=0.0,
        )



import numpy as np
import pandas as pd


def process_cn_raster(
    hsg,
    parcels,
    season,
    composite_landuse,
    parcels_ids,
    cn_table,
    nodata=-9999,
):
    """Creates the CN-map for a given year and season.

    Current implementation executes three steps:

    - Generate a CNmaxID based on CN type id (1-11, associated to crop), contour
      plowing measures (0/1/2), hydrological conditions ({0,1->3},
      {unknown, poor to good}), contour_id (0/1).
    - Couple CN-values based in CNmaxID and season for all parcels (based on parcel id)
    - Coupled CN-values 99, 99 and 1 to land-uses river (-1) , infrastructure (-2) and
      water (-5).


    Parameters
    ----------
    hsg: numpy.ndarray
        Raster that holds values from 1 to 4 according to hydrological soil group
        (A, B, C and D), see [1]_ and [2]_, for more info, see documentation in function
        :attr: pywatemsedem.catchment.Catchment.hydrological_soil_group`.
    parcels: geopandas.GeoDataFrame
        Parcels polygon data with columns:

        - *NR* (int): unique id, should be the same as in parcels_ids (i.e. parcels_ids
          is a rasterized version of the parcels dataframe).
        - "CN_type_id" (int): crop specific CN id (1-11).
        - "HydroCond" (str): Either "Poor", "Fair" or "Good".

        Optional columns:

        - *gewasrest* (int): are the 'gewasresten' yes (1) or no (0).
        - *contour* (int): is contour plowing applied yes (1) or no (0).
        - *ntkerende* (int): is there non-plowing tillage yes (1) or no (0).
        - *verslemp* (int): has silting occurred yes (1) or no (0).

    season: str
        Either "fall", "spring", "summer", "winter"

    composite_landuse: numpy.ndarray
        A composite land-use raster. Typically, it defines the landuse from
        non-agricultural pixels.
    parcels_ids: numpy.ndarray
        A parcels id raster, holding unique id's per parcel.
    cn_table: pandas.DataFrame
        Table holding CNmax values per hydrological soil group and per CNmaxID.
    nodata: float, default -9999
        Nodata value

    Returns
    -------
    arr_cn: numpy.ndarray
        Raster with values between 1 (low CN, low run-off) and 100 (high CN,
        high run-off)


    Notes
    -----
    1. The CNtype_id is crop/cover-specific. Note that for general land-uses reported
       in the base landuse map (no crop info present), CN is set equal to a predefined
       CNmax:

       - forest: CN = CNmax is based on CNmaxID 9030
       - grassland: CN = CNmax is based on CNmaxID 5020
       - grass: CN = CNmax is based on CNmaxID 6000

    2. CN is set equal to 99 for infrastructure, rivers and pools (no infiltration)

    3. The contour_id is computed by:

        - contour_id = 1 if CNtype_id in {2, 3, 4} and contour = 0
        - contour_id = 2 if CNtype_id in {2, 3, 4} and contour = 1
        - contour_id = 0 else

    4. The HydroCond_id is computed by:

        - HydroCond_id = 1 if HydroCond is equal to 'Poor'.
        - HydroCond_id = 2 if HydroCond is equal to 'Fair'.
        - HydroCond_id = 3 if HydroCond is equal to 'Good'.
        - HydroCond_id = 0 if HydroCond is equal to NULL.

    5. Gewasrest_id is equal to gewasrest (1 or 0)

    6. When the CNtype_id is equal to NULL, CNtype_id is set to 4,
       HydroCond_id to 1 and contour_id to 1.


    References
    ----------
    .. [1] NRCS, 2004. Hydrologic soil-cover complexes (chapter 9).
    .. [2] NRCS, 2009. Hydrologic soil groups (chapter 7).
    """
    if season not in ["summer", "spring", "winter", "fall"]:
        msg = "Season should be 'summer', 'spring', 'winter' and 'fall'."
        raise ValueError(msg)

    df = parcels.copy()
    # standard set contour-source-oriented measures off
    if "contour" not in df:
        df["contour"] = 0
    # maak een contour_id per perceel op basis van het CN-type (1-11) en
    # de contouroptie van de percelen
    df["contour_id"] = 0  # standaardwaarde
    cond = df["CNtype_id"].isin([2, 3, 4]) & df["contour"].isin(
        [0]
    )  # CNtype 2,3 of 4 en geen contourploegen
    df["contour_id"] = np.where(cond, 1, df["contour_id"])  # contour_id krijgt 1 (SR)
    cond = df["CNtype_id"].isin([2, 3, 4]) & df["contour"].isin(
        [1]
    )  # CNtype 2,3 of 4 en contourploegen
    df["contour_id"] = np.where(cond, 2, df["contour_id"])  # contour_id krijgt 2 (C)

    # hydrologische toestand van de bodem op het perceel
    df["HydroCond_id"] = 0  # standaardwaarde
    df["HydroCond_id"] = np.where(df["HydroCond"] == "Poor", 1, df["HydroCond_id"])
    df["HydroCond_id"] = np.where(df["HydroCond"] == "Fair", 2, df["HydroCond_id"])
    df["HydroCond_id"] = np.where(df["HydroCond"] == "Good", 3, df["HydroCond_id"])

    # gewasrest
    if "gewasrest" not in df:
        df["gewasrest"] = 0
    df["gewasrest_id"] = np.where(df["gewasrest"] == 1, 1, 0)

    # zet alle percelen waar geen gegevens (onbekende teelt) voor zijn
    # op onbekend
    cond = df["CNtype_id"].isnull()
    df["CNtype_id"] = np.where(cond, 4, df["CNtype_id"])
    df["HydroCond_id"] = np.where(cond, 1, df["HydroCond_id"])
    df["contour_id"] = np.where(cond, 1, df["contour_id"])

    df["CNmaxID"] = (
        df["CNtype_id"].astype(int).map(str)
        + df["contour_id"].map(str)
        + df["HydroCond_id"].map(str)
        + df["gewasrest_id"].map(str)
    )
    df["CNmaxID"] = df["CNmaxID"].astype(int)
    df = pd.merge(df, cn_table, how="left", on="CNmaxID")

    # correctiefactoren cover en verslemping
    # geen gegevens, dus 0 --> geen correctie bij berekening CN

    # Voorlopige invulling seizoensafhankelijkheid CN -->
    # cijfers gebaseerd op bedekkingsgraad van 15 veelvoorkomende gewassen
    # (cover.txt), behoorlijk hoge bedekkingsgraden in winter te wijten aan
    # groot aandeel grasland binnen de Vlaamse percelen
    cover_seasons = np.array([40.1919, 52.3333, 80.7869, 47.4675])

    if season == "winter":
        df["cover"] = cover_seasons[0]
    elif season == "spring":
        df["cover"] = cover_seasons[1]
    elif season == "summer":
        df["cover"] = cover_seasons[2]
    elif season == "fall":
        df["cover"] = cover_seasons[3]

    df["verslemp"] = 0

    arr_soil_classes = np.unique(
        hsg.astype(int)
    ).tolist()  # de arr_soil_classes aanwezig in het modelgebied
    arr_soil_classes = [cl for cl in arr_soil_classes if cl != nodata]
    # voor elk perceel wordt voor elke soil_class de CN-waarde berekend
    c2 = 3
    for soil_class in arr_soil_classes:
        cn_min = f"CNmin_{int(soil_class)}"
        cn_max = f"CNmax_{int(soil_class)}"
        c1 = f"c1_{int(soil_class)}"
        cn_model = f"CNmodel_{int(soil_class)}"
        df[cn_min] = 4.2 * df[cn_max] / (10 - 0.058 * df[cn_max])
        df[c1] = df[cn_max] - df[cn_min]
        df[cn_model] = (
            df[cn_max] - ((df["cover"] / 100) * df[c1]) + ((df["verslemp"] / 5) * c2)
        )
        # correctie niet kerende bodembewerking
        if "ntkerend" not in df:
            df["ntkerend"] = 0
        df[cn_model] = np.where(df["ntkerend"] == 1, df[cn_model] - 1, df[cn_model])

    # reclass perceelskaart --> kan beter! df.apply?
    arr_cn = np.ones(parcels_ids.shape) * nodata

    for row in df.itertuples():
        rstval = getattr(row, "NR")
        for soil_class in arr_soil_classes:
            newval = getattr(row, f"CNmodel_{int(soil_class)}")
            cond = np.logical_and(
                (parcels_ids == rstval),
                (hsg.astype(int) == soil_class),
            )
            arr_cn = np.where(cond, newval, arr_cn)

    # rivier, infrastructuur,  die niet op de landbouwperceelskaart staan.
    # waar geen percelen zijn moeten we opvullen op basis van landgebruik.
    arr_cn = np.where((composite_landuse == -1), 99, arr_cn)
    arr_cn = np.where(
        (composite_landuse == -2), 99, arr_cn
    )  # wegen en water krijgen CN 99
    arr_cn = np.where(
        (composite_landuse == -5), 0, arr_cn
    )  # vijvers en poelen krijgen CN 0

    # bossen, weiden en grasland die niet op de landbouwperceelskaart staan.
    # waar geen percelen zijn moeten we opvullen op basis van landgebruik.

    cn_table.index = cn_table["CNmaxID"]
    for soil_class in arr_soil_classes:
        cond = arr_cn == nodata
        z = cn_table[f"CNmax_{int(soil_class)}"].loc[9030]
        arr_cn[(cond) & (composite_landuse == -3) & (hsg.astype(int) == soil_class)] = z
        cond = arr_cn == nodata
        z = cn_table[f"CNmax_{int(soil_class)}"].loc[5020]
        arr_cn[(cond) & (composite_landuse == -4) & (hsg.astype(int) == soil_class)] = z
        cond = arr_cn == nodata
        z = cn_table[f"CNmax_{int(soil_class)}"].loc[6000]
        arr_cn[(cond) & (composite_landuse == -6) & (hsg.astype(int) == soil_class)] = z

    arr_cn[(parcels_ids == -9999) & (composite_landuse > 0)] = np.mean(
        arr_cn[(arr_cn >= 0) & (arr_cn <= 100)]
    )
    # !!!!!!! VOORLOPIGE FIX OM TE HOGE CN WAARDEN TE VERMIJDEN
    arr_cn[arr_cn > 100] = np.mean(arr_cn[(arr_cn >= 0) & (arr_cn <= 100)])
    arr_cn[arr_cn == nodata] = np.mean(arr_cn[(arr_cn >= 0) & (arr_cn <= 100)])

    return arr_cn, df



import shutil

import pytest
from conftest import ini_file

from pywatemsedem.io.ini import add_field, modify_field


class TestModifyField:
    def test_modify(self, tmp_path):
        """Test basic func"""
        modify_field(ini_file, "User Choices", "max kernel", 3)

    def test_wrong_section(self, tmp_path):
        """Expect error when the section does not exist"""
        with pytest.raises(KeyError) as excinfo:
            modify_field(ini_file, "User Choice", "max kernel", 3)
        assert "Section 'User Choice' does not exist" in str(excinfo.value)

    def test_wrong_key(self, tmp_path):
        """Expect error when the to-modify-key does not exist"""
        with pytest.raises(KeyError) as excinfo:
            modify_field(ini_file, "User Choices", "max kernelzz", 3)
        assert "Key 'max kernelzz' does not exist" in str(excinfo.value)


class TestAddField:
    def test_add(self, tmp_path):
        """Test basic func"""
        shutil.copy(ini_file, tmp_path / "test.ini")
        add_field(tmp_path / "test.ini", "User Choices", "max kernelzzz", 3)

    def test_wrong_section(self, tmp_path):
        """Expect error when the section does not exist"""
        with pytest.raises(KeyError) as excinfo:
            add_field(ini_file, "User Choice", "max kernel", 3)
        assert "Section 'User Choice' does not exist" in str(excinfo.value)

    def test_key_exists(self, tmp_path):
        """Expect error when the to-add-key already exists"""
        with pytest.raises(KeyError) as excinfo:
            add_field(ini_file, "User Choices", "max kernel", 3)
        assert "Key 'max kernel' already exist in" in str(excinfo.value)



import logging

# Standard libraries
import subprocess
import warnings
from copy import deepcopy
from functools import wraps

import geopandas as gpd
import numpy as np
import pandas as pd

from pywatemsedem.cfactor import create_cfactor_cnws
from pywatemsedem.cn import process_cn_raster
from pywatemsedem.geo.rasters import RasterMemory
from pywatemsedem.geo.utils import nearly_identical, saga_intersection
from pywatemsedem.io.folders import ScenarioFolders
from pywatemsedem.io.ini import IniFile
from pywatemsedem.ktc import create_ktc_cnws
from pywatemsedem.parcelslanduse import ParcelsLanduse, get_source_landuse

from .buffers import (
    filter_outlets_in_arr_extension_id,
    process_buffer_outlets,
    process_buffers_in_river,
)
from .errors import (
    attribute_continuous_value_error,
    attribute_discrete_value_error,
    missing_attribute_error_in_vct,
)
from .grasstrips import create_grassstrips_cnws
from .templates import InputFileName
from .tools import format_forced_routing, zip_folder

inputfilename = InputFileName()
DEFAULT_CFACTOR = 0.37
logger = logging.getLogger(__name__)


def valid_vct_endpoints(func):
    """Check if endpoints vector are defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.choices.dict_model_options["Include sewers"] == 1:
            if self._vct_endpoints is None:
                msg = (
                    "Please define endpoints line vector (see "
                    "vct_endpoints-property) or set 'Include sewers' in "
                    "'dict_model_options' to 0!!"
                )
                raise IOError(msg)

            elif "SewerInletEff" not in self.choices.dict_variables:
                msg = "Please define a 'SewerInletEff' in 'dict_variables'."
                raise KeyError(msg)
            return func(self, *args, **kwargs)
        else:
            msg = "Please define 'Include sewers' in 'dict_model_options' to 1."
            raise IOError(msg)

    return wrapper


def valid_vct_parcels(func):
    """Check if parcels vector are defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self._vct_parcels is None:
            msg = "Please define parcels polygon vector (see vct_parcels-property)!"
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_source_measure(func):
    """Check if source measure vector are defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self._vct_source_measure is None:
            msg = (
                "Please define source measure polygon vector (see "
                "vct_source_measure-property)!"
            )
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_grass_strips(func):
    """Check if grass strips vector is defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.choices.dict_ecm_options["UseGras"] == 1:
            if self._vct_grass_strips is None:
                msg = "No grass strips defined, but option 'UseGras' equal to 1."
                raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_buffers(func):
    """Check if buffers vector is defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.choices.dict_ecm_options["Include buffers"] == 1:
            if self._vct_buffers is None:
                msg = "No buffers defined, but option 'Include buffers' equal to 1."
                warnings.warn(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_outlets(func):
    """Check if outlet vector is defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self._vct_outlets is None:
            msg = "Please define outlets point vector (see vct_outlets-property)!"
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_dtm(func):
    """Check if you have defined a DTM."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.catchm.dtm is None:
            msg = "Please first define a DTM!"
            raise IOError(msg)

        return func(self, *args, **kwargs)

    return wrapper


def valid_kfactor(func):
    """Check if you have defined a K-factor raster."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.catchm.kfactor is None:
            msg = "Please first define a K-factor raster!"
            raise IOError(msg)

        return func(self, *args, **kwargs)

    return wrapper


def valid_pfactor(func):
    """Check if you have defined a P-factor raster."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.catchm.pfactor is None:
            msg = "Please first define a P-factor raster!"
            raise IOError(msg)

        return func(self, *args, **kwargs)

    return wrapper


def valid_cnws_input(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.choices.version == "CN-WS":
            if self.cn is None:
                msg = (
                    "Please create a valid CN-raster with *prepare_cnws_model_input*"
                    " for a WaTEM/SEDEM run."
                )
                raise IOError(msg)

        if self.ktc is None:
            if self.choices.dict_model_options["UserProvidedKTC"] == 1:
                msg = (
                    "Please create a valid kTC-raster with *prepare_cnws_model_input*."
                )
                raise IOError(msg)

        if self.cfactor is None:
            msg = (
                "Please create a valid c-factor-raster with "
                "*prepare_cnws_model_input*."
            )
            raise IOError(msg)

        if self.composite_landuse is None:
            msg = (
                "Please define a WaTEM/SEDEM-landuse-raster with "
                "*prepare_cnws_model_input*."
            )
            raise IOError(msg)

        return func(self, *args, **kwargs)

    return wrapper


def valid_ini(func):
    """Check if you have defined a valid ini-file."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """wrapper"""
        if self.ini is None:
            msg = "Please define an ini-file with *create_ini_file*."
            raise IOError(msg)

        return func(self, *args, **kwargs)

    return wrapper


class CNWSException(Exception):
    """Exception from WaTEM/SEDEM pre- and postprocessing scripts"""


class Scenario:
    """Construct a new Scenario instance

    The catchment class holds all dynamic information for a scenario, for a given year.
    The class considers all data of which the content is user-option-dependent (i.e.
    :class:`pywatemsedem.userchoices.UserChoices`.

    - *vct_parcels*: polygon vector of parcels with obligated columns 'LANDUSE',
      'C_crop' and 'CODE'.
    - *vct_grass_strips*: polygon vector of grass stips with obligated columns
      'scale_ktc' and 'width'.
    - *vct_buffers*: polygon vector of grass stips with obligated columns 'buffercap'
      (CN), 'hdam' (CN), 'hknijp' (CN), 'dknijp' (CN), 'qcoef' (CN), 'boverl' (CN),
      'eff' (WS).
    - *vct_bufferoutlets*: #TODO
    - *vct_outlets*: point vector.
    - *vct_ditches*: #TODO
    - *vct_conductive_dams*: #TODO
    - *vct_force_routing*: line vector.
    - *vct_endpoints*: line vector, with optional columns 'efficiency' and 'type_id'.
    """

    def __init__(self, catchm, year, scenario_nr, userchoices):
        """Generate scenario instance with catchment, year, scenario number and user
        choices.

        Parameters
        ----------
        catchm: pywatemsedem.core.catchment.Catchment
            Instance of :class:`pywatemsedem catchment <pywatemsedem.catchment.Catchment>`
            containing the catchment characteristics.
        year: int
            Simulation year
        scenario_nr: int
            Identifier number of the scenario.
        userchoices : pywatemsedem.core.userchoices.UserChoices
            Containing the pywatemsedem model settings.
        """
        # Init factories from catchm instance
        self.catchm = catchm
        self.rp = self.catchm.rp

        # Initalize
        self.vector_factory = self.catchm.vector_factory
        self.raster_factory = self.catchm.raster_factory

        # properties
        self._grass_strips = None  # raster
        self._vct_grass_strips = None  # vector
        self._vct_buffers = None  # vector
        self._vct_bufferoutlets = None  # vector
        self._vct_ditches = None
        self._vct_condictive_dams = None
        self._vct_source_measures = None
        self._vct_endpoints = None
        self._endpoints = None
        self._endpoints_id = None
        self._vct_force_routing = None
        self._vct_outlets = None
        self._outlets = None
        self._vct_parcels = None

        # derivate properties
        self._ktc = None
        self._cn = None
        self._cn_table = None
        self._cfactor = None
        self._composite_landuse = None

        # assign scenario number and user choices
        self.scenario_nr = scenario_nr
        self.year = year
        self.choices = deepcopy(userchoices)
        self.rst_outlet = None
        self.ini = None

        # initialisation functionalities
        self.temporal_resolution()
        self.create_folder_structure()
        self.copy_data_layers_to_scenario_folder()

    def create_folder_structure(self):
        """#TODO"""

        self.scenario_folder_init = (
            self.catchm.folder.home_folder / f"scenario_" f"{self.scenario_nr}"
        )

        self.sfolder = ScenarioFolders(self.catchm.folder, self.scenario_nr, self.year)
        self.sfolder.create_all()

    @valid_kfactor
    @valid_dtm
    @valid_pfactor
    def copy_data_layers_to_scenario_folder(self):
        """#TODO"""
        self.catchm.kfactor.write(
            self.sfolder.cnwsinput_folder / inputfilename.kfactor_file
        )
        self.catchm.dtm.write(
            self.sfolder.cnwsinput_folder / inputfilename.dtm_file, nodata=-99999
        )
        self.catchm.pfactor.write(
            self.sfolder.cnwsinput_folder / inputfilename.pfactor_file
        )
        self.catchm.adjacent_edges.to_csv(
            self.sfolder.cnwsinput_folder / inputfilename.adjacentedges_file,
            sep="\t",
            index=False,
        )
        self.catchm.up_edges.to_csv(
            self.sfolder.cnwsinput_folder / inputfilename.upedges_file,
            sep="\t",
            index=False,
        )

        if self.choices.dict_model_options["River Routing"] == 1:
            self.choices.dict_output["Output per river segment"] = 1
            self.catchm.routing.write(
                self.sfolder.cnwsinput_folder / inputfilename.routing_file
            )
        # if self.choices.dict_output["Output per river segment"] == 1:
        self.catchm.segments.write(
            self.sfolder.cnwsinput_folder / inputfilename.segments_file
        )

        self.catchm.mask.write(self.sfolder.cnwsinput_folder / inputfilename.mask_file)

    def temporal_resolution(self):
        """Calculates for which years and seasons the scenario needs data.

        Based on the defined choices in the
        :py:class:`CNWS.UserChoices <pywatemsedem.CNWS.UserChoices>` 'begin_jaar',
        'begin_maand' and, in case of CNWS, 'Endtime model'.
        """
        if (self.choices.version == "WS") or self.choices.version == "Only Routing":
            self.season = "spring"
        else:
            if self.choices.dict_variables["begin_maand"] in [1, 2, 3]:
                self.season = "winter"
            elif self.choices.dict_variables["begin_maand"] in [4, 5, 6]:
                self.season = "spring"
            elif self.choices.dict_variables["begin_maand"] in [7, 8, 9]:
                self.season = "summer"
            elif self.choices.dict_variables["begin_maand"] in [10, 11, 12]:
                self.season = "fall"

    @property
    def vct_parcels(self):
        """Getter parcels polygon vector"""

        return self._vct_parcels

    @vct_parcels.setter
    def vct_parcels(self, vector_input):
        """Assign parcels polygon vector

        The parcels vector should be polygon vector. Should contain a definition of
        land-use (column "LANDUSE"):

            - *-4*: grass land
            - *-3*: forest
            - *-2*: infrastructure (farms)
            - *-9999*: agricultural land

        Should contain a definition of crop code (column "CODE").

        Should contain a definition of the crop C-factor (column 'C_crop', see
        :ref:`here <watemsedem:cmap>`.):

            - *[0,1]*: C-factor for crop.
            - *NULL*: no C-factor defined.

        Should contain a defiction of the C-reduction (column 'C_reduction'), used
        in case of source-oriented measures.

        Can contain a 'NR' column which is the identification ID of the individual
        parcel.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Polygon vector

            - *LANDUSE* (int): landuse value (-4: grass land, -3: forest,
               -2: infrastructure (farms), -9999: agricultural land).
            - *C_crop* (float): C-factor for crop, valid for considered time period
              ([0,1], NULL-values allowed).
            - *NR* (int, optional): id.
            - *C_reduction* (float): C-reduction values,  C-factor is reduced with this
              percentage when source-oriented measures are used.

        Notes
        -----
        The C-value for crops are defined as the C-factor that is valid as value for
        the coupled crop given a time period (e.g. average C-factor for potato for one
        year, or average C-factor for potato for the month April).
        """
        self._vct_parcels = self.vector_factory(
            vector_input, "Polygon", allow_empty=True
        )

        missing_attribute_error_in_vct(
            self._vct_parcels.geodata,
            "Parcels",
            {"C_crop", "CODE", "LANDUSE", "C_reduct"},
        )
        attribute_continuous_value_error(
            self._vct_parcels.geodata, "Parcels", "C_crop", lower=0, upper=1
        )
        attribute_continuous_value_error(
            self._vct_parcels.geodata, "Parcels", "C_reduct", lower=0, upper=1
        )
        attribute_discrete_value_error(
            self._vct_parcels.geodata,
            "Parcels",
            "LANDUSE",
            {-9999, -2, -3, -4, -5},
            classes={"agriculture", "infrastructure", "forest", "grass land", "water"},
        )

        if "NR" not in self._vct_parcels.geodata.columns:
            self._vct_parcels.geodata["NR"] = range(
                0, len(self._vct_parcels.geodata), 1
            )

        self._vct_parcels.geodata["LANDUSE"] = self._vct_parcels.geodata[
            "LANDUSE"
        ].astype(float)

    @property
    def parcels(self):
        """Get parcels raster

        Raster contains id's (defined in "NR", if not defined: according to
        id sequence in vector). The parcel raster is limited by the value of 32767
        (int16). If values are higher, then these are limited to 32767.

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Int16 raster with values:

            - *>0 and <=32767*: parcel id
            - *-9999*: nodata
        """
        if self.parcels_ids is not None:
            arr = self.parcels_ids.arr.copy()
            if np.any(arr > 32767):
                msg = (
                    "Parcels raster has values higher than the maximum allowed number "
                    "for WaTEM/SEDEM definition (i.e. 32767). Setting values above "
                    "32767 to 32767."
                )
                warnings.warn(msg)
            arr = np.where(arr > 32767, 32767, arr)
            arr = arr.astype(np.int16)

            return self.raster_factory(arr)
        else:
            return None

    @property
    def parcels_ids(self):
        """Getter rasterized parcels polygon vector

        Raster contains id's (defined in "NR", if not defined: according to
        id sequence in vector).

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster, else None
            Float64 raster with values:

            - *>0*: parcel id
            - *-9999*: nodata
        """
        if self._vct_parcels is not None:
            arr = self._vct_parcels.rasterize(
                self.catchm.mask_raster, self.rp.epsg, col="NR", gdal=True
            )
            return self.raster_factory(arr)
        else:
            return None

    @property
    def parcels_landuse(self):
        """Getter rasterized parcels polygon vector

        Raster contains landuse (defined in column "LANDUSE")

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster, else None
            Float64 raster with values:

            - *-4*: grass land
            - *-3*: forest
            - *-2*: infrastructure (farms)
            - *-9999*: agricultural land

        Notes
        -----
        If all parcels have a landuse code of -9999, then a None is returned
        (no addition to landuse).
        """
        if self._vct_parcels is not None:
            arr = self.vct_parcels.rasterize(
                self.catchm.mask_raster, self.rp.epsg, col="LANDUSE", gdal=True
            )
            if np.all(arr == self.rp.nodata):
                return None
            else:
                return self.raster_factory(arr)
        else:
            return None

    @property
    def vct_grass_strips(self):
        """Getter grass strips polygon vector"""

        return self._vct_grass_strips

    @vct_grass_strips.setter
    def vct_grass_strips(self, vector_input):
        """Assign grass strips polygon vector

        Grass strips are appointed the value -6 in the WaTEM/SEDEM parcels landuse
        raster. In addition, C-factors and kTC-values are assigned to grass strips
        pixels according to their width, see also
        :func:`pywatemsedem.cfactor.create_cfactor_cnws` and
        :func:`pywatemsedem.ktc.create_ktc_cnws`

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Polygon vector

            - *width* (int): meter width of grass strip (shortest side!)
            - *scale_ktc* (int): either 0 (scale or 1), if not kTC is not scaled, then
              the ktc_low (see :class:`pywatemsedem.userchoices.UserChoices`) is used,
              see :func:`pywatemsedem.ktc.scale_ktc_gdf_grass_strips`

        Notes
        -----
        Grass strips are defined by a length and a width, the width is considered to be
        the shortest side, whereas the length is considerd to be the longest side.
        """
        self._vct_grass_strips = self.vector_factory(vector_input, "Polygon")
        missing_attribute_error_in_vct(
            self._vct_grass_strips.geodata, "Grass strips", {"width", "scale_ktc"}
        )
        attribute_continuous_value_error(
            self._vct_grass_strips.geodata, "Grass strips", "width", lower=0
        )
        attribute_discrete_value_error(
            self._vct_grass_strips.geodata,
            "Grass strips",
            "scale_ktc",
            {0, 1},
            classes={
                "do not scale ktc grass strip according to width",
                "scale ktc according to width",
            },
        )

        if "NR" not in self._vct_grass_strips.geodata.columns:
            self._vct_grass_strips._geodata["NR"] = np.arange(
                0, len(self._vct_grass_strips._geodata), 1
            )
        self._vct_grass_strips.geodata["NR"] = self._vct_grass_strips.geodata[
            "NR"
        ].astype(float)
        arr = self.vct_grass_strips.rasterize(
            self.catchm.mask_raster,
            self.rp.epsg,
            col="NR",
            nodata=-9999,
            gdal=True,
        )
        self._grass_strips = self.raster_factory(arr)

    @property
    @valid_vct_grass_strips
    def grass_strips(self):
        """Grass strips raster getter"""
        if self.choices.dict_ecm_options["UseGras"] == 0:

            msg = (
                "'UseGras' in 'dict_ecm_options' is equal to 0. Will not include grass"
                " strips."
            )
            warnings.warn(msg)
            return None
        else:
            return self._grass_strips

    @grass_strips.setter
    def grass_strips(self, raster_input):
        """Grass strips raster setter

        Parameters
        ----------
        raster_input: Pathlib.Path, str or numpy.ndarray
            See :func:`pywatemsedem.catchment.raster_factory`.
        """
        self._grass_strips = self.raster_factory(raster_input)

    @property
    def vct_buffers(self):
        """Getter buffers polygon vector"""
        return self._vct_buffers

    @vct_buffers.setter
    def vct_buffers(self, vector_input):
        """Assign buffers polygon vector

        Buffers are considered in WaTEM/SEDEM as a raster defined by an id (outlet) and
        extension id (buffer extent, i.e. number of pixels surrounding the outlet). In
        addition, a number of properties of the buffers are assigned (i.e. 'buffercap',
        'hdam', 'hknijp', 'dknijp', 'qcoef', 'boverl', 'eff').

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Polygon vector, with columns

            - *id* (WS)
            - *buffercap* (CN)
            - *hdam* (CN)
            - *hknijp* (CN)
            - *dknijp* (CN)
            - *qcoef* (CN)
            - *boverl* (CN)
            - *eff* (WS)

            For an explanation of the columns, see :ref:`here <watemsedem:buffermap>`
        """

        req_col = {
            "buffercap",
            "hdam",
            "hknijp",
            "dknijp",
            "qcoef",
            "boverl",
            "eff",
        }
        self._vct_buffers = self.vector_factory(vector_input, "Polygon")
        missing_attribute_error_in_vct(self._vct_buffers.geodata, "Buffers", req_col)
        attribute_continuous_value_error(
            self._vct_buffers.geodata, "Buffers", "eff", lower=0, upper=100
        )
        self._vct_buffers.geodata["id"] = np.arange(
            1.0, len(self.vct_buffers.geodata) + 1, 1
        )

        if self.choices.dict_ecm_options["Include buffers"] == 0:
            msg = (
                "Include buffers' in 'dict_ecm_options' is equal to 0. Will not "
                "include buffers."
            )
            warnings.warn(msg)

    @property
    def vct_bufferoutlets(self):
        """Getter infrastructure polygon vector"""
        return self._vct_bufferoutlets

    @vct_bufferoutlets.setter
    def vct_bufferoutlets(self, vector_input):
        """Assign buffer outlets vector

        #TODO

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            See :func:`pywatemsedem.catchment.vector_factory`.
        """
        self._vct_bufferoutlets = self.vector_factory(vector_input, "Polygon")

    @property
    @valid_vct_buffers
    def buffers_exid(self):
        """Getter buffer extension Id raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
        """
        arr = self.vct_buffers.rasterize(
            self.catchm.mask_raster, 31370, "buf_exid", gdal=True
        )

        return self.raster_factory(arr)

    @property
    def bufferoutlet(self):
        """Getter buffer outlet raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
        """
        if self.vct_bufferoutlets is not None:
            self.vct_bufferoutlets.geodata = process_buffer_outlets(
                self.vct_bufferoutlets.geodata, self.vct_buffers.geodata
            )
            arr = np.where(self.catchm.vct_river.arr == -1, 0, self.bufferoutlet)
        else:
            arr = np.zeros(self.catchm.dtm.arr.shape)

        return self.raster_factory(arr)

    @property
    @valid_vct_buffers
    def buffers(self):
        """Getter for buffer

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster

        Notes
        -----
        1. Overlapping buffers with rivers (complete overlap) are removed.
        2. Buffer outlet vector can be used to define the buffer outlets, if not used
           then the pixel with the lowest DTM height value is used as outlet for a
           buffer.
        """
        arr = None
        if self.choices.dict_ecm_options["Include buffers"] == 0:
            msg = (
                "Option 'Include buffers' in erosion control measure options is 0, "
                "returning None"
            )
            warnings.warn(msg)
        elif self.vct_buffers is None:
            self.choices.dict_ecm_options["Include buffers"] = 0
        else:
            self.vct_buffers.geodata = assign_buffer_id_to_df_buffer(
                self.vct_buffers.geodata
            )
            if self.catchm.vct_river is not None:
                arr, self.vct_buffers.geodata = process_buffers_in_river(
                    self.vct_buffers.geodata,
                    self.buffers_exid.arr,
                    self.catchm.river.arr,
                    self.rp.gdal_profile["nodata"],
                )
                if self.vct_buffers.geodata.empty:  # if all buffers are removed
                    msg = (
                        "All buffers pixels overlap with the river, removing and not "
                        "including buffers in simulation."
                    )
                    logger.warning(msg)
                    self.choices.dict_ecm_options["Include buffers"] = 0
                    arr = None
            if arr is not None:
                arr = filter_outlets_in_arr_extension_id(
                    self.vct_buffers.geodata, arr, self.catchm.dtm.arr, None
                )
                arr = np.where(arr == self.catchm.rp.nodata, 0, arr).astype("int16")
                arr = self.raster_factory(arr, flag_mask=False)

        return arr

    @property
    def vct_ditches(self):
        """Getter ditches lines vector"""
        return self._vct_ditches

    @vct_ditches.setter
    def vct_ditches(self, vector_input):
        """Setter ditches lines vector

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            See :func:`pywatemsedem.catchment.vector_factory`.
        """
        self._vct_ditches = self.vector_factory(vector_input, "Polygon")

    @property
    def vct_conductive_dams(self):
        """Getter conductive dams vector"""
        return self._vct_condictive_dams

    @vct_conductive_dams.setter
    def vct_conductive_dams(self, vector_input):
        """Setter conductive dams lines vector

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            See :func:`pywatemsedem.catchment.vector_factory`.
        """
        self._vct_condictive_dams = self.vector_factory(vector_input, "Polygon")

    def conductive_dams(self):
        """Getter conductive dams

        Returns
        -------
        pywatemsedem.geo.raster.AbstractRaster
        """
        arr = self.vct_conductive_dams.rasterize(
            self.catchm.mask_raster,
            self.rp.epsg,
            convert_lines_to_direction=True,
            gdal=True,
        )
        if self.buffers is not None:
            # if there is a buffer on the same pixel of a ditch, remove the ditch
            arr = np.where(self.buffers.arr != 0, 0, arr)
        if self.catchm.vct_river is not None:
            arr = np.where(self.river.arr == -1, 0, arr)

        return self.raster_factory(arr)

    def ditches(self):
        """Getter ditches

        Returns
        -------
        pywatemsedem.geo.raster.AbstractRaster
        """
        arr = self.vct_ditches.rasterize(
            self.catchm.mask_raster,
            self.rp.epsg,
            convert_lines_to_direction=True,
            gdal=True,
        )
        if self.buffers is not None:
            # if there is a buffer on the same pixel of a ditch, remove the ditch
            arr = np.where(self.buffers.arr != 0, 0, arr)
        if self.catchm.vct_river is not None:
            arr = np.where(self.river.arr == -1, 0, arr)

        return self.raster_factory(arr)

    @property
    def vct_outlets(self):
        """Getter outlets vector

        Returns
        -------
        object
        """
        return self._vct_outlets

    @vct_outlets.setter
    def vct_outlets(self, vector_input):
        """Set vector outlet point

        With this setter, one can define the outlet point of a catchment. This outlet
        is reflected in the output of WaTEM/SEDEM (see
        :ref:`here <watemsedem:totalsedimenttxt>`).

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Point vector, with optional column 'NR'.
        """
        self._vct_outlets = self.vector_factory(vector_input, "Point")

        if len(self._vct_outlets.geodata) == 0:
            msg = r"Defined outlets are not in catchment."
            logger.warning(msg)
            self._vct_outlets = None
        else:
            if "NR" not in self.vct_outlets.geodata:
                self._vct_outlets.geodata["NR"] = np.arange(
                    1.0, len(self.vct_outlets.geodata) + 1, 1
                )
            self._vct_outlets.geodata["NR"] = self._vct_outlets.geodata["NR"].astype(
                float
            )

    @property
    @valid_vct_outlets
    def outlets(self):
        """Getter outlets

        Returns
        -------
        pywatemsedem.geo.raster.AbstractRaster
        """

        arr = self.vct_outlets.rasterize(
            self.catchm.mask_raster, self.rp.epsg, col="NR", gdal=True
        )
        self._outlets = self.raster_factory(arr)
        self._outlets.arr = self._outlets.arr.astype(np.int16)
        return self._outlets

    @property
    def vct_force_routing(self):
        """#todo

        Returns
        -------

        """
        return self._vct_force_routing

    @vct_force_routing.setter
    def vct_force_routing(self, vector_input):
        """Appoint force routing

        Parameters
        ----------
        vector_input: str | pathlib.Path | geopandas.GeoDataFrame
            See :func:`pywatemsedem.catchment.vector_factory`.
        """
        vct_forced_routing = self.vector_factory(
            vector_input, "LineString", allow_empty=True
        )
        self._vct_force_routing = format_forced_routing(
            vct_forced_routing.geodata,
            self.catchm.rp.gdal_profile["minmax"],
            self.catchm.rp.resolution,
        )

    @property
    def vct_endpoints(self):
        """Getter endpoints lines vector"""
        return self._vct_endpoints

    @vct_endpoints.setter
    def vct_endpoints(self, vector_input):
        """Assign endpoints line vector

        Endpoints are defined as sink objects, in which sediment is captured and not
        transported further. The endpoints should be line vector.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Line vector with optional column

            - *type_id* (int): describes the different classes of endpoints (e.g. 1:
              sewers, 2: ditches, ...)
            - *efficiency* (float): sediment trap efficiency (%) .

        Notes
        -----
        1. In the WaTEM/SEDEM pywatemsedem engine, endpoints are named "sewers", see also
           :ref:`here <watemsedem:sewermapfile>`. Note that in pywatemsedem, type of endpoints can
           be defined (ditches, sewers).
        """
        self._vct_endpoints = self.vector_factory(
            vector_input, "LineString", allow_empty=True
        )

        if "type_id" not in self._vct_endpoints.geodata.columns:
            msg = "No 'type_id' assigned, assuming one type of endpoint with id 1."
            warnings.warn(msg)
            self._vct_endpoints["type_id"] = 1
        else:
            if np.any(self._vct_endpoints.geodata["type_id"].isnull()):
                msg = "Please define a 'type_id' for every record."
                raise ValueError(msg)

        if "efficiency" not in self._vct_endpoints.geodata.columns:
            self._vct_endpoints.geodata["efficiency"] = np.nan
        else:
            attribute_continuous_value_error(
                self._vct_endpoints.geodata, "Endpoints", "efficiency", 0, 1
            )

        self._vct_endpoints.geodata["efficiency"] = self._vct_endpoints.geodata[
            "efficiency"
        ].astype(np.float64)

    @property
    @valid_vct_endpoints
    def endpoints_id(self):
        """Getter endpoints id raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Float64 raster with values:

            - *0*: no endpoint
            - *not equal to 0*: id.
        """
        arr_id = self.vct_endpoints.rasterize(
            self.catchm.mask_raster, 31370, col="type_id", gdal=True
        )
        arr_id[arr_id == self.rp.nodata] = 0

        return self.raster_factory(arr_id, flag_mask=False, flag_clip=False)

    @property
    @valid_vct_endpoints
    def endpoints(self):
        """Getter endpoints efficiency raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Float64 raster with values in [0,1] efficiency (in decimal).

        Notes
        -----
        Id's equal to one are converted to 0 when
        self.choices.dict_model_options["OnlyInfraSewers"] is equal to 1.
        """
        cond = self._vct_endpoints.geodata["efficiency"].isnull()
        if np.any(cond):
            # in decimals
            self._vct_endpoints.geodata.loc[cond, "efficiency"] = float(
                self.choices.dict_variables["SewerInletEff"]
            )
            msg = (
                "The efficiency is not defined for all sewer line strings, assigning"
                f" 'SewerInletEff'-value defined in user choices 'variables' "
                f"({self.choices.dict_variables['SewerInletEff']*100} %)."
            )
            warnings.warn(msg)

        self.vct_endpoints.geodata["efficiency"] = self.vct_endpoints.geodata[
            "efficiency"
        ].astype(float)
        arr = self.vct_endpoints.rasterize(
            self.catchm.mask_raster,
            31370,
            nodata=self.rp.nodata,
            col="efficiency",
            gdal=False,
        )

        arr[arr == self.rp.nodata] = 0
        if self.choices.dict_model_options["OnlyInfraSewers"] == 1:
            cond = (self.catchm.infrastructure.arr != -2) & (self.endpoints_id.arr == 1)
            cond_mask = (cond) & (self.catchm.mask.arr != 0)
            arr[cond_mask] = 0
        arr[self.catchm.river.arr == -1] = 0
        return self.raster_factory(arr, flag_mask=False, flag_clip=False)

    @property
    def cn_table(self):
        """Set/getter CN-table

        Returns
        -------
        pandas.DataFrame
            with columns:

            - CNmaxID (int): unique identifier, compiled from CN type id (1-11,
              associated to crop), contour plowing measures (0/1/2), hydrological
              conditions ({0,1->3}, {unknown, poor to good}), contour_id (0/1).
            - CNmax_1 (int): CNmax for soil class 1
            - CNmax_2 (int): CNmax for soil class 2
            - CNmax_3 (int): CNmax for soil class 3
            - CNmax_4 (int): CNmax for soil class 4
        """
        return self._cn_table

    @cn_table.setter
    def cn_table(self, df):
        """Set/getter CN-table

        Returns
        -------
        pandas.DataFrame
            with columns:

            - CNmaxID (int): unique identifier, compiled from CN type id (1-11,
              associated to crop), contour plowing measures (0/1/2), hydrological
              conditions ({0,1->3}, {unknown, poor to good}), contour_id (0/1).
            - CNmax_1 (int): CNmax for soil class 1
            - CNmax_2 (int): CNmax for soil class 2
            - CNmax_3 (int): CNmax for soil class 3
            - CNmax_4 (int): CNmax for soil class 4
        """
        if not {"CNmaxID", "CNmax_1", "CNmax_2", "CNmax_3", "CNmax_4"}.issubset(df):
            msg = (
                "Dataframe should have columns 'CNmaxID', 'CNmax_1', 'CNmax_2', "
                "'CNmax_3', 'CNmax_4'."
            )
            raise IOError(msg)
        self._cn_table = df

    @property
    def composite_landuse(self):
        """Getter infrastructure polygon vector"""
        return self._composite_landuse

    @composite_landuse.setter
    def composite_landuse(self, raster_input):
        """Assign parcels polygon vector

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            See :func:`pywatemsedem.catchment.vector_factory`.
        """
        self._composite_landuse = self.raster_factory(
            raster_input, flag_mask=False, flag_clip=False
        )

    @property
    def cfactor(self):
        """Getter cfactor 3-D raster"""
        return self._cfactor

    @cfactor.setter
    def cfactor(self, raster_input):
        """Setter cfactor 3-D raster

        Parameters
        ----------
        raster_input:
            3-D raster: x,y raster for every season.
        """
        raster = self.raster_factory(raster_input)
        raster.arr = np.where(raster.arr == raster.rp.nodata, 0, raster.arr)
        self._cfactor = raster

    @property
    def ktc(self):
        """Getter ktc 3-D raster"""

        return self._ktc

    @ktc.setter
    def ktc(self, raster_input):
        """Setter ktc 3-D raster

        Parameters
        ----------
        raster_input:
            3-D raster: x,y raster for every season.
        """
        self._ktc = self.raster_factory(raster_input)

    @property
    def cn(self):
        """Getter cn 3-D raster"""
        return self._cn

    @cn.setter
    def cn(self, raster_input):
        """Setter cn 3-D raster

        Parameters
        ----------
        raster_input:
            3-D raster: x,y raster for every season.
        """
        self._cn = self.raster_factory(raster_input, flag_clip=False, flag_mask=False)

    def prepare_grass_strips(self, maximize_grass_strips=False):
        """Prepare grass strips array

        Parameters
        ----------
        maximize_grass_strips: bool, default False

        """
        if self.grass_strips is not None:
            arr_parcels = None if self.parcels is None else self.parcels.arr
            arr_grass_ids, arr_grass = create_grassstrips_cnws(
                self.grass_strips.arr,
                self.catchm.river.arr,
                self.catchm.infrastructure.arr,
                self.rp.rasterio_profile,
                arr_parcels=arr_parcels,
                expand_grass_strips=maximize_grass_strips,
            )

            # write_ids
            grass_ids = RasterMemory(arr_grass_ids, self.rp)
            grass_ids.write(
                self.sfolder.scenarioyear_folder
                / (
                    f"grasstripsID_{self.year}_{self.catchm.name}_s"
                    f"{self.scenario_nr}.tif"
                ),
                format="tiff",
                dtype=np.int16,
            )
            # grass
            self.grass_strips = arr_grass
            self.grass_strips.write(
                self.sfolder.scenarioyear_folder
                / (f"gras_{self.year}_{self.catchm.name}_s{self.scenario_nr}.tif"),
                format="tiff",
                dtype=np.int16,
            )

    def create_perceelskaart_cnws(
        self,
        rivers,
        water,
        infrastructure,
        landuse,
        mask,
        yearfolder,
        rp,
        grass_strips=None,
        parcels=None,
        landuse_parcels=None,
    ):
        """Create CN_WS perceelskaart according to format of :ref:`here <watemsedem:prcmap>`.

        Parameters
        ----------
        rivers: numpy.ndarray
            River raster, should only contain nodata-value and -1 (river), see
            :ref:`here <watemsedem:prcmap>`.
        water: numpy.ndarray
            Water raster, should only contain nodata-value and -3 (), see
            :ref:`here <watemsedem:prcmap>`.
        infrastructure: numpy.ndarray
            Infrastructure raster, should only contain -2 (paved) and -7 (non-paved)
            and nodata-value. #TODO: check if we can only use -7
        landuse: numpy.ndarray
            Landuse raster. This raster should be formatted according to the composite
            landuse / WaTEM/SEDEM parcels raster (see :ref:`here <watemsedem:prcmap>`).
        mask: numpy.ndarray
            Mask raster,
        yearfolder: pathlib.Path
            #TODO: try to ditch this
        rp: pywatemsedem.geo.rasterproperties.RasterProperties
            Raster properties, see :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
        grass_strips: numpy.ndarray
            Grass strips raster, can only contain nodata-value and -6 (see
            :ref:`here <watemsedem:prcmap>`).
        parcels: numpy.ndarray
            Parcels ids raster, can only containt nodata-value and >0 (see
            :ref:`here <watemsedem:prcmap>`)

        Notes
        -----
        The infrastructur value -7 does not follow the definition of
        :ref:`here <watemsedem:prcmap>`. This value is converted in subfunctionalities of
        this function to -2.
        """
        # parcels)
        # TODO
        if self.parcels is not None:
            maxprc_id = np.max(self.parcels.arr)
        else:
            maxprc_id = 1

        # landuse
        landuse_core = get_source_landuse(
            landuse,
            maxprc_id,
            rp.rasterio_profile,
            mask,
            yearfolder,
            self.catchm.name,
        )
        pl = ParcelsLanduse(
            landuse_core,
            rivers,
            water,
            infrastructure,
            mask,
            rp.nodata,
            landuse_parcels=landuse_parcels,
            parcels=parcels,
            grass_strips=grass_strips,
        )
        arr = pl.create_parcels_landuse_raster()
        arr = np.where(arr == -7, -2, arr)  # aardewegen infstructuur maken
        # safety check, fill last empty gaps with 32767
        arr[(mask == 1) & (arr == 0)] = 32767
        composite_landuse = np.where(mask == 1, arr, 0).astype("float64")

        return composite_landuse

    def prepare_cnws_model_input(self, maximize_grass_strips=False):
        """Create model input for a scenario

        Parameters
        ----------
        maximize_grass_strips: bool, default False
            See :func:`pywatemsedem.scenario.create_perceelskaart_cnws`
        """
        logger.info("Aanmaken van alle nodige modelinput...")
        self.prepare_grass_strips(maximize_grass_strips=maximize_grass_strips)

        arr_parcels_lu = (
            None if self.parcels_landuse is None else self.parcels_landuse.arr
        )

        # todo: how could we fix following lines?
        if self.grass_strips is None:
            arr_grass_strips = None
        else:
            arr_grass_strips = self.grass_strips.arr
            arr_grass_strips[arr_grass_strips != self.grass_strips.rp.nodata] = -6

        arr_water = None if self.catchm.vct_water is None else self.catchm.water.arr

        arr_parcels = None if self.parcels is None else self.parcels.arr
        self.composite_landuse = self.create_perceelskaart_cnws(
            self.catchm.river.arr,
            arr_water,
            self.catchm.infrastructure.arr,
            self.catchm.landuse,
            self.catchm.mask.arr_bin,
            self.sfolder.scenarioyear_folder,
            self.catchm.rp,
            grass_strips=arr_grass_strips,
            parcels=arr_parcels,
            landuse_parcels=arr_parcels_lu,
        )

        self.composite_landuse.write(
            self.sfolder.cnwsinput_folder / inputfilename.parcelmosaic_file,
            dtype=np.int32,
        )

        if self.choices.version != "Only Routing":

            self.update_seasonal_data(
                self.composite_landuse,
                self.catchm.river,
                self.catchm.infrastructure,
                self.catchm.landuse,
                self.catchm.mask,
                vct_parcels=self.vct_parcels,
                vct_grass_strips=self.vct_grass_strips,
            )

        if self.choices.dict_model_options["Manual outlet selection"] == 1:
            self.outlets.write(
                self.sfolder.cnwsinput_folder / inputfilename.outlet_file
            )
        if self.choices.version != "Only Routing":
            if self.choices.dict_model_options["Calibrate"] == 1:
                for key in [
                    "Write sediment export",
                    "Write water erosion",
                    "Output per river segment",
                ]:
                    self.choices.dict_output[key] = 0
        if (self.choices.dict_ecm_options["Include buffers"] == 1) & (
            self.buffers is not None
        ):
            self.buffers.write(
                self.sfolder.cnwsinput_folder / inputfilename.buffers_file
            )

        if self.choices.dict_ecm_options["Include ditches"] == 1:
            self.ditches.write(
                self.sfolder.cnwsinput_folder / inputfilename.ditches_file
            )

        if self.choices.dict_ecm_options["Include dams"] == 1:
            self.conductive_dams.write(
                self.sfolder.cnwsinput_folder / inputfilename.conductivedams_file
            )

        if self.choices.dict_model_options["FilterDTM"] == 1:
            msg = "Filtering DTM within boundaries of parcels."
            logger.info(msg)
            self.catchm.dtm.filter()
            self.catchm.dtm.write(
                self.sfolder.cnwsinput_folder / inputfilename.dtm_file
            )

        if self.choices.dict_model_options["Include sewers"] == 1:
            if self.endpoints is not None:
                self.endpoints.write(
                    self.sfolder.cnwsinput_folder / inputfilename.endpoints_file,
                    format="idrisi",
                    dtype=np.float64,
                )
                self.endpoints_id.write(
                    self.sfolder.cnwsinput_folder / inputfilename.endpoints_id_file,
                    format="idrisi",
                    dtype=np.float64,
                )

    def update_seasonal_data(
        self,
        composite_landuse,
        river,
        infrastructure,
        landuse,
        mask,
        vct_parcels=None,
        vct_grass_strips=None,
    ):
        """Update seasonal data for prepare_cnws_model_input

        Parameters
        ----------
        year: int
            Simulation year.
        cn_table:
            File path to CN table.
        """
        # C-factor
        vct_grass_strips, cfactor = create_cfactor_cnws(
            river,
            infrastructure,
            landuse,
            mask,
            self.sfolder,
            vct_parcels=vct_parcels,
            vct_grass_strips=vct_grass_strips,
            use_source_oriented_measures=bool(
                self.choices.dict_ecm_options["UseTeelttechn"]
            ),
        )
        self.cfactor = cfactor
        self.cfactor.write(self.sfolder.cnwsinput_folder / inputfilename.cfactor_file)

        # CN
        if self.choices.version == "CN-WS":
            self.cn, gdf = process_cn_raster(
                self.catchm.hydrological_soil_group.arr,
                self.vct_parcels.geodata,
                self.season,
                self.composite_landuse.arr,
                self.parcels_ids.arr,
                self.cn_table,
                self.rp.nodata,
            )
            gdf.to_file(self.sfolder.cnwsinput_folder / "CN_table_parcels.shp")
            self.cn.write(self.sfolder.cnwsinput_folder / inputfilename.cn_file)
        # kTC
        if self.choices.dict_model_options["UserProvidedKTC"] == 1:
            out, vct_grass_strips = create_ktc_cnws(
                composite_landuse,
                cfactor,
                mask,
                self.choices.dict_variables["ktc low"],
                self.choices.dict_variables["ktc high"],
                self.choices.dict_variables["ktc limit"],
                self.sfolder,
                grass=vct_grass_strips,
            )
            # write results
            self.ktc = out
            self.ktc.write(self.sfolder.cnwsinput_folder / inputfilename.ktc_file)
            # write grass strips
            if vct_grass_strips is not None:
                vct_grass_strips.write(
                    self.sfolder.scenarioyear_folder
                    / f"gras_{self.year}_{self.catchm.name}_s{self.scenario_nr}.shp"
                )

    @valid_cnws_input
    def create_ini_file(self):
        """Creates an ini-file for the scenario"""
        logger.info("Aanmaken ini-file...")
        self.ini = self.sfolder.cnwsinput_folder / "inifile.ini"
        ini = IniFile(
            self.choices,
            self.choices.version,
            self.sfolder.cnwsinput_folder,
            self.sfolder.cnwsoutput_folder,
        )
        ini.add_model_information()
        ini.add_working_directories()
        ini.add_files()
        ini.add_user_choices()
        ini.add_output_maps()
        ini.add_variables(
            self.vct_buffers.geodata if self.vct_buffers is not None else None,
            self.vct_force_routing,
            self.catchm.vct_tubed_river,
        )
        ini.write(self.ini)

    @valid_ini
    def run_model(self, cnws_binary):
        """Run the WaTEM/SEDEM model

        Parameters
        ----------
        cnws_binary : str
            Name of CN_WS pascal compiled executable.

        """
        logger.info(f"Modeling scenario {self.scenario_nr}")
        try:
            cmd_args = [cnws_binary, str(self.ini)]
            run = subprocess.run(cmd_args, capture_output=True)
            for line in run.stdout.splitlines():
                logger.info(line.decode("utf-8"))
            logger.info("Modelrun finished!")
        except subprocess.CalledProcessError as e:
            msg = "Failed to run CNWS-model"
            logger.exception(msg)
            logger.exception(e.cmd)
            raise IOError(e)

    def zip(self):

        zip_folder(self.sfolder.scenario_folder)


def remove_known_grass_strips_from_parcels_vct_saga(parcels, vct_grass_strips):
    """Remove parcels with overlap above 75%

    Remove the parcels in vct_parcels that overlap with grass strips
    in vct_grass_strips if the overlap for both features is at least 75%.

    The function updates the vct_parcels input layer

    The amount of intersection or overlap between a parcel is caluclated as
    area_intersect(grass strips, parcel)/area(parcel).

    Parameters
    ----------
    parcels: geopandas.GeoDataFrame
        Can have any column, should have geometries
    vct_grass_strips: str or pathlib.Path
        File path of parcel shapefile
    """
    cond = (vct_grass_strips is not None) and (parcels is not None)
    fname_temp = vct_grass_strips.parents[0] / "temp_intersect.shp"
    fname_parcels = vct_grass_strips.parents[0] / "temp_parcels.shp"

    if cond:
        # check if NR is in vct_parcels
        if "NR" not in parcels:
            parcels["NR"] = np.arange(1, len(parcels) + 1, 1)
        parcels.to_file(fname_parcels)
        saga_intersection(
            str(fname_parcels),
            str(vct_grass_strips),
            fname_temp,
        )

        # intersect
        parcels_intersection = gpd.read_file(fname_temp)

        # couple considered parcels from this year and prior year (only consider
        # largest intersect)
        parcels_intersection["area_inter"] = parcels_intersection["geometry"].area

        # find largest interesecting parcel and select only those
        parcels_intersection["area_inter_max"] = parcels_intersection.groupby("NR")[
            "area_inter"
        ].transform(np.max)
        cond = (
            parcels_intersection["area_inter_max"] == parcels_intersection["area_inter"]
        )
        parcels_intersection = parcels_intersection[cond]

        # load parcels
        parcels.loc[:, "area"] = parcels["geometry"].area.values

        parcels = parcels.merge(
            parcels_intersection[["NR", "area_inter"]],
            on="NR",
            how="left",
        )

        # normalize area intersction  with area of parcel of considered year
        parcels["norm_over"] = parcels["area_inter"] / parcels["area"] * 100

        # define condition
        cond = (parcels["norm_over"].isnull()) | (parcels["norm_over"] <= 75)
        parcels = parcels.loc[cond]
        parcels = parcels.drop(
            columns=["area_inter_max", "norm_over", "area_inter"], errors="ignore"
        )

    return parcels


def convert_grass_strips_to_agricultural_fields(parcels, vct_grass_strips):
    """Convert known grass strips to agricultural fields in the parcels shape file.

    If a parcel in vct_parcels intersects for 75% or more with a grass strip in
    vct_grass_strips, the attributes of the parcel are changed: "LANDUSE" is set
    to -9999 (agriculture), "GWSCOD_H" is set to "9999" (unknown crop) and
    "C-factor" is set to 0.37. This function updates the input vct_parcels.

    The amount of intersection or overlap between a parcel is caluclated as
    area_intersect(grasstri, parcel)/area(parcel).

    Parameters
    ----------
    parcels: geopandas.GeoDataFrame
        Can have any column, should have geometries

    vct_grass_strips: str or pathlib.Path
        File path of parcel shapefiles

    Returns
    -------
    parcels: geopandas.GeoDataFrame
        With added column:

        - *default_cfactor* (bool): convert to default C-factor (and landuse).

    """

    if vct_grass_strips is not None and parcels is not None:
        if "C_factor" not in parcels:
            parcels["C_factor"] = np.nan
        gdf_grass = gpd.read_file(vct_grass_strips)
        matches = parcels.geometry.apply(lambda x: nearly_identical(gdf_grass, x))
        matches = matches.unstack().reset_index(0, drop=True).dropna()
        gdf_grass_matched = gdf_grass.reindex(index=matches.values)
        gdf_grass_matched.index = matches.index
        gdf_grass_matched["grasstrip"] = 1
        gdf_grass_matched = gdf_grass_matched[["grasstrip"]]
        parcels = pd.merge(
            parcels,
            gdf_grass_matched,
            how="left",
            left_index=True,
            right_index=True,
        )
        parcels.loc[parcels["grasstrip"] == 1, "LANDUSE"] = -9999
        parcels.loc[parcels["grasstrip"] == 1, "GWSCOD_H"] = 9999
        parcels.loc[parcels["grasstrip"] == 1, "C_factor"] = 0.37

    return parcels


def assign_buffer_id_to_df_buffer(df):
    """Assign buffer id and extention id to buffer dataframe.

    Parameters
    ----------
    df: pandas.DataFrame
        Buffer dataframe with columns

    Returns
    ----------
    df: pandas.DataFrame
        Updated with:

        - *buf_id* (int): buffer id.
        - *buf_exid* (int): buffer extension id.
    """
    df["buf_id"] = range(1, df.shape[0] + 1)
    df["buf_exid"] = df["buf_id"] + 2**14

    return df


def add_tillage_technical_measures_to_parcels(gdf_parcels, gdf_tillage_technical):
    """Add crop technical measures to parcels vector data.

    An overlap between the target parcel polygons and source technical polygons is
    computed to determine for which parcels a tillage technical measure should be
    applied.

    Parameters
    ----------
    gdf_parcels: geopandas.GeoDataFrame
        Parcels polygon vector
    gdf_tillage_technical: geopandas.GeoDataFrame
        Tillage technical polygon vector
    overlap: int
        Minimal required overlap between polygons to select the implementation of a
        tillage technical measure to be applied for the parcel.
    """

    matches = gdf_parcels.geometry.apply(
        lambda x: nearly_identical(gdf_tillage_technical, x, 0.75)
    )
    matches2 = matches.unstack().reset_index(0, drop=True).dropna()
    df_teelttech_matched = gdf_tillage_technical.reindex(index=matches2.values)
    df_teelttech_matched.index = matches2.index

    df_teelttech_matched = df_teelttech_matched[["category"]]

    df_merged = pd.merge(
        gdf_parcels, df_teelttech_matched, how="left", left_index=True, right_index=True
    )
    df_merged.loc[df_merged["category"] == "ntkerend", "ntkerend"] = 1
    df_merged.loc[df_merged["category"] == "drempels", "drempels"] = 1
    df_merged.loc[df_merged["category"] == "contourzaaien", "contour"] = 1
    df_merged.loc[df_merged["category"] == "gewasrest", "gewasrest"] = 1
    df_merged.loc[df_merged["category"] == "groenbedekker", "groenbedek"] = 1
    df_merged.loc[df_merged["category"] == "default", "ntkerend"] = 1
    df_merged.drop("category", axis=1, inplace=True)

    return df_merged



from pathlib import Path

import pandas as pd

from pywatemsedem.geo.utils import grid_difference
from pywatemsedem.io.modeloutput import load_total_sediment_file
from pywatemsedem.tools import zip_folder

from .postprocess import PostProcess


class DeltaScenario:
    """Construct a DeltaScenario instance with two instances of PostProcess

    Parameters
    ----------
    pp1: pywatemsedem.postproces.PostProcess
        First PostProcess instance of scenario used for the comparison.
    pp2: pywatemsedem.postproces.PostProcess
        Second PostProcess instance of scenario used for the comparison.
    resmap:
        Folder path used to write results to

    Examples
    --------
    Define inputs
        >>> pp1 = PostProcess(molenbeek, 1, 2019, 20, 31370)
        >>> pp2 = PostProcess(molenbeek, 2, 2019, 20, 31370)

    Run
        >>> delta_scenario = DeltaScenario(pp1,pp2)
        >>> delta_scenario.comparison_total_sediment_file()
        >>> ...
    """

    def __init__(self, pp1, pp2, resmap=None):
        """

        Parameters
        ----------
        pp1: pywatemsedem.core.postprocess.PostProcess
            See class:`pywatemsedem.postprocess.PostProcess`
        pp2: pywatemsedem.core.postprocess.PostProcess
            See class:`pywatemsedem.postprocess.PostProcess`
        resmap: str or pathlib.Path, default None
            Folder path of results map
        """
        if self.check_input(pp1):
            self.pp1 = pp1
        if self.check_input(pp2):
            self.pp2 = pp2
        if resmap is None:
            self.resmap = (
                Path.cwd()
                / pp1.catchment_name
                / f"compare_scenario{pp1.scenario}_{pp2.scenario}"
            )
        else:
            self.resmap = Path(resmap)
        if not self.resmap.exists():
            self.resmap.mkdir(parents=True, exist_ok=True)

        self.check_rasterprop()

    def check_input(self, pp):
        """Check the type of input given to DeltaScenario class

        Parameters
        ----------
        pp : pywatemsedem.postproces.PostProcess
            A PostProcess instance of scenario used for the comparison.
        """
        if not isinstance(pp, PostProcess):
            msg = (
                f"{pp} is not an instance of the class PostProcess, "
                f"see ~:class:pywatemsedem.postprocess.PostProcess"
            )
            raise IOError(msg)
        else:
            return True

    def check_rasterprop(self):
        """Check if the raster properties are equal for both scenario's"""
        rstparams1 = self.pp1.rstparams
        rstparams2 = self.pp2.rstparams
        if rstparams1 == rstparams2:
            return True
        else:
            return False

    def zip_files(self):
        """Zip output folder of deltascenario"""

        zip_folder(self.resmap)

    def comparison_total_sediment_file(self):
        """Compare the values in the total sediments file

        Returns
        -------
        df: pandas.DataFrame
            Holding results of summary values for run for scenario 1 and 2,
            and the difference between those values.

            - *sc1*: values for general statistics scenario 1
            - *sc2*: values for general statistics scenario 2
            - *difference*: difference between statistics scenario 1 and 2

        """
        df_1 = load_total_sediment_file(self.pp1.files["txt_total_sediment"])
        df_2 = load_total_sediment_file(self.pp2.files["txt_total_sediment"])

        df = pd.DataFrame(index=df_1.keys(), columns=["sc1", "sc2", "difference"])
        df["sc1"] = df_1.values()
        df["sc2"] = df_2.values()
        df["difference"] = df["sc1"] - df["sc2"]

        return df

    def difference_grids(self):
        """Make difference grids for a number of files"""
        difference_grids = [
            "rst_sediout",
            "rst_sediin",
            "rst_sediexport",
            "rst_watereros",
        ]

        for grid in difference_grids:
            rst_1 = self.pp1.files[grid]
            rst_2 = self.pp2.files[grid]
            rst_out = self.resmap / (grid + "_difference.rst")
            grid_difference(rst_1, rst_2, rst_out)



import pytest
from conftest import geodata
from geopandas import GeoDataFrame
from numpy import array

from pywatemsedem.geo.factory import Factory
from pywatemsedem.geo.valid import PywatemsedemInputError


class TestFactory:

    """Test class for factory

    Parameters
    ----------
    nodata: -9999
    epsg_code: 31370 (Lambert)
    resolution: 20 (m)

    """

    nodata = -9999
    epsg_code = 31370
    resolution = 20

    @pytest.mark.saga
    def test_raster_mask(self, tmp_path):
        """Test factory with raster

        ----------
        tmp_path
        """
        f = Factory(self.resolution, self.epsg_code, self.nodata, tmp_path)
        assert f.create_mask(geodata.rst_mask)

    def test_vector_mask(self, tmp_path):
        """Test factory with vector

        ----------
        tmp_path
        """
        f = Factory(self.resolution, self.epsg_code, self.nodata, tmp_path)
        assert f.create_mask(geodata.catchment)

    @pytest.mark.saga
    def test_error_no_mask(self, tmp_path):
        """Test error when no mask is created

        Parameters
        ----------
        tmp_path
        """
        f = Factory(self.resolution, self.epsg_code, self.nodata, tmp_path)

        # feed raster to vector factory
        with pytest.raises(PywatemsedemInputError) as excinfo:
            f.vector_factory(geodata.rst_mask, "Polygon")
        assert (
            f"First create a mask with {Factory.create_mask.__name__}-function"
            in str(excinfo.value)
        )

    @pytest.mark.saga
    def test_error_input_to_factory(self, tmp_path):
        """Test error in input to factory

        Parameters
        ----------
        tmp_path
        """
        f = Factory(self.resolution, self.epsg_code, self.nodata, tmp_path)
        f.create_mask(geodata.rst_mask)

        # feed raster to vector factory
        with pytest.raises(IOError) as excinfo:
            f.vector_factory(geodata.rst_mask, "Polygon")
        assert (
            f"Input vector file '{geodata.rst_mask}' should be a valid "
            f"vector file (e.g. ESRI shape file)." in str(excinfo.value)
        )

        # feed numpy array to vector factory
        with pytest.raises(IOError) as excinfo:
            f.vector_factory(array([]), "Polygon")
        assert "Input vector should be a geopandas GeoDataFrame or vector file." in str(
            excinfo.value
        )

        # feed vector file to raster factory
        with pytest.raises(IOError) as excinfo:
            f.raster_factory(geodata.catchment)
        assert (
            f"Input raster file '{geodata.catchment}' should be a valid raster file "
            f"(e.g. IDRISI raster, geotiff, SAGA-GRID, ..)" in str(excinfo.value)
        )

        # feed geopandas dataframe to raster factory
        with pytest.raises(IOError) as excinfo:
            f.raster_factory(GeoDataFrame())
        assert "Input raster should be a numpy array or raster file," in str(
            excinfo.value
        )
